#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.optimisation.operators import LinearOperator
from cil.framework import BlockGeometry, BlockDataContainer
from cil.optimisation.operators import FiniteDifferenceOperator


class SymmetrisedGradientOperator(LinearOperator):

    r''' The symmetrised gradient is the operator, :math:`E`, defined by :math:`E: V \rightarrow W` where `V` is `BlockGeometry` and  `W` is the range of the Symmetrised Gradient and

        .. math::

            E(v) = 0.5  ( \nabla v + (\nabla v)^{T} ) \\


    In 2 dimensions, let :math:`v(x,y)=(v_1(x,y),v_2(x,y))` which gives

        .. math::

            \nabla v =\left( \begin{matrix}
                \partial_{x} v_1 & \partial_x v_2\\
                \partial_{y}v_1 & \partial_y v_2
            \end{matrix}\right)

    and thus

        .. math::

            E(v) = 0.5  ( \nabla v + (\nabla v)^{T} )
            =\left( \begin{matrix}
                \partial_{x} v_1 & 0.5  (\partial_{y} v_1 + \partial_{x} v_2) \\
                0.5  (\partial_{x} v_1 + \partial_{y} v_2) & \partial_{y} v_2
            \end{matrix}\right)

    Parameters
    ----------
    domain_geometry: `BlockGeometry` with shape (2,1) or (3,1)
        Set up the domain of the function.
    bnd_cond: str, optional, default :code:`Neumann`
        Boundary condition either :code:`Neumann` or :code:`Periodic`
    correlation: str, optional, default :code:`Channel`
        Correlation either :code:`SpaceChannel` or :code:`Channel`

    '''

    CORRELATION_SPACE = "Space"
    CORRELATION_SPACECHANNEL = "SpaceChannels"

    def __init__(self, domain_geometry, bnd_cond = 'Neumann', **kwargs):

        self.bnd_cond = bnd_cond
        self.correlation = kwargs.get('correlation',SymmetrisedGradientOperator.CORRELATION_SPACE)

        tmp_gm = len(domain_geometry.geometries)*domain_geometry.geometries


        # Define FD operator. We need one geometry from the BlockGeometry of the domain
        self.FD = FiniteDifferenceOperator(domain_geometry.get_item(0), direction = 0,
                             bnd_cond = self.bnd_cond)

        if domain_geometry.shape[0]==2:
            self.order_ind = [0,2,1,3]
        else:
            self.order_ind = [0,3,6,1,4,7,2,5,8]

        super(SymmetrisedGradientOperator, self).__init__(
                                          domain_geometry=domain_geometry,
                                          range_geometry=BlockGeometry(*tmp_gm))


    def direct(self, x, out=None):

        r'''Returns :math:`E(v) = 0.5 * ( \nabla v + (\nabla v)^{T} )`

        Parameters:
        -------------

        x: BlockDataContainer
        out: BlockDataContainer, default None
            If out is not None the output of direct will be filled in out, otherwise a new object is instantiated and returned.
        '''

        if out is None:

            tmp = []
            for i in range(self.domain_geometry().shape[0]):
                for j in range(x.shape[0]):
                    self.FD.direction = i
                    tmp.append(self.FD.adjoint(x.get_item(j)))

            tmp1 = [tmp[i] for i in self.order_ind]

            res = [0.5 * sum(x) for x in zip(tmp, tmp1)]

            return BlockDataContainer(*res)

        else:

            ind = 0
            for i in range(self.domain_geometry().shape[0]):
                for j in range(x.shape[0]):
                    self.FD.direction = i
                    self.FD.adjoint(x.get_item(j), out=out[ind])
                    ind+=1
            out1 = BlockDataContainer(*[out[i] for i in self.order_ind])
            out.fill( 0.5 * (out + out1) )
            return out

    def adjoint(self, x, out=None):
        r'''Returns the adjoint of the symmetrised gradient operator

        Parameters:
        -------------

        x: BlockDataContainer
        out: BlockDataContainer, default None
            If out is not None the output of adjoint will be filled in out, otherwise a new object is instantiated and returned.
        '''

        if out is None:

            tmp = [None]*self.domain_geometry().shape[0]
            i = 0

            for k in range(self.domain_geometry().shape[0]):
                tmp1 = 0
                for j in range(self.domain_geometry().shape[0]):
                    self.FD.direction = j
                    tmp1 += self.FD.direct(x[i])
                    i+=1
                tmp[k] = tmp1
            return BlockDataContainer(*tmp)


        else:

            tmp = self.domain_geometry().allocate()
            i = 0
            for k in range(self.domain_geometry().shape[0]):
                tmp1 = 0
                for j in range(self.domain_geometry().shape[0]):
                    self.FD.direction = j
                    self.FD.direct(x[i], out=tmp[j])
                    i+=1
                    tmp1+=tmp[j]
                out[k].fill(tmp1)
            return out




#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
import functools
import logging
import sys

import numpy as np

from cil.framework import (DataContainer, ImageGeometry, ImageData, VectorGeometry, AcquisitionData,
                           AcquisitionGeometry, BlockGeometry, VectorData)
from cil.framework.labels import ImageDimension, AcquisitionDimension

from testclass import CCPiTestClass
from utils import initialise_tests

log = logging.getLogger(__name__)
initialise_tests()


def aid(x):
    """returns the memory block address of an array"""
    return x.as_array().__array_interface__['data'][0]


class TestDataContainer(CCPiTestClass):
    def create_DataContainer(self, X,Y,Z, value=1):
        a = value * np.ones((X, Y, Z), dtype='float32')
        #print("a refcount " , sys.getrefcount(a))
        ds = DataContainer(a, False, ['X', 'Y', 'Z'])
        return ds

    def test_creation_nocopy(self):
        shape = 2, 3, 4, 5
        size = np.prod(shape)
        a = np.arange(size)
        #print("a refcount " , sys.getrefcount(a))
        a = np.reshape(a, shape)
        #print("a refcount " , sys.getrefcount(a))
        ds = DataContainer(a, False, ['X', 'Y', 'Z', 'W'])
        #print("a refcount " , sys.getrefcount(a))
        self.assertEqual(id(a), id(ds.array))
        self.assertEqual(ds.dimension_labels, ('X', 'Y', 'Z', 'W'))

    def testGb_creation_nocopy(self):
        X, Y, Z = 8, 64, 64
        a = np.ones((X, Y, Z), dtype='float32')
        #print("a refcount " , sys.getrefcount(a))
        ds = DataContainer(a, False, ['X', 'Y', 'Z'])
        #print("a refcount " , sys.getrefcount(a))
        self.assertEqual(sys.getrefcount(a), 3)
        ds1 = ds.copy()
        self.assertNotEqual(aid(ds), aid(ds1))
        ds1 = ds.clone()
        self.assertNotEqual(aid(ds), aid(ds1))

    def test_ndim(self):
        x_np = np.arange(0, 60).reshape(3,4,5)
        x_cil = DataContainer(x_np)
        self.assertEqual(x_np.ndim, x_cil.ndim)
        self.assertEqual(3, x_cil.ndim)

    def test_DataContainer_equal(self):
        array = np.linspace(-1, 1, 32, dtype=np.float32).reshape(4, 8)
        data = DataContainer(array)
        data1 = data.copy()

        # Check two identical DataContainers are equal
        self.assertTrue((data == data1).all())
        # Check it works for comparing DataContainer with numpy array
        self.assertTrue((data == data1.as_array()).all())

        # # Check two different DataContainers are not equal
        data1 += 1
        self.assertFalse((data ==  data1).all())

    def test_AcquisitionData_equal(self):
        array = np.linspace(-1, 1, 32, dtype=np.float32).reshape(4, 8)
        geom = AcquisitionGeometry.create_Parallel3D().set_panel((8, 4)).set_channels(1).set_angles([1])
        data = AcquisitionData(array, geometry=geom)

        data1 = data.copy()

        # Check two identical AcquisitionData are equal
        self.assertTrue(data == data1)
        # Check it works for comparing AcquisitionData with numpy array
        self.assertTrue(data == data1.as_array())

        # # Check two different AcquisitionData are not equal
        data1 += 1
        self.assertFalse(data ==  data1)

        # Check the equality of two AcquisitionData with different shapes
        data_different_shape = data.copy()
        data_different_shape.array = data_different_shape.array.reshape(8, 4)

        self.assertFalse(data == data_different_shape)

        # Check the equality of two AcquisitionData with different dtypes
        data_different_dtype = data.copy()
        data_different_dtype.array = data_different_dtype.array.astype(np.float64)
        self.assertFalse(data == data_different_dtype)


        # Check the equality of two AcquisitionData with different labels
        data_different_labels = data.copy()
        print(data_different_labels.geometry.dimension_labels)
        data_different_labels.geometry.set_labels([AcquisitionDimension("ANGLE"), AcquisitionDimension("CHANNEL") ])
        self.assertFalse(data == data_different_labels)

    def test_ImageData_equal(self):
        array = np.linspace(-1, 1, 32, dtype=np.float32).reshape(4, 8)
        geom = ImageGeometry(voxel_num_x=8, voxel_num_y=4)
        data = ImageData(array, geometry=geom)

        data1 = data.copy()

        # Check two identical ImageData are equal
        self.assertTrue(data == data1)
        # Check it works for comparing ImageData with numpy array
        self.assertTrue(data == data1.as_array())

        # # Check two different ImageData are not equal
        data1 += 1
        self.assertFalse(data ==  data1)

        # Check the equality of two ImageData with different shapes
        data_different_shape = data.copy()
        data_different_shape.array = data_different_shape.array.reshape(8, 4)

        self.assertFalse(data == data_different_shape)

        # Check the equality of two ImageData with different dtypes
        data_different_dtype = data.copy()
        data_different_dtype.array = data_different_dtype.array.astype(np.float64)
        self.assertFalse(data == data_different_dtype)


        # Check the equality of two ImageData with different labels
        data_different_labels = data.copy()
        data_different_labels.geometry.set_labels([ImageDimension("VERTICAL"), ImageDimension("HORIZONTAL_X")])
        self.assertFalse(data == data_different_labels)


    def testInlineAlgebra(self):
        X, Y, Z = 8, 16, 32
        a = np.ones((X, Y, Z), dtype='float32')
        b = np.ones((X, Y, Z), dtype='float32')
        ds = DataContainer(a, False, ['X', 'Y', 'Z'])
        ds += 2
        # self.assertEqual(ds.as_array()[0][0][0], 3.)
        np.testing.assert_array_almost_equal(ds.as_array(), 3 * b)
        ds -= 2
        # self.assertEqual(ds.as_array()[0][0][0], 1.)
        np.testing.assert_array_almost_equal(ds.as_array(), b)

        ds *= 2
        # self.assertEqual(ds.as_array()[0][0][0], 2.)
        np.testing.assert_array_almost_equal(ds.as_array(), b * 2)

        ds /= 2
        # self.assertEqual(ds.as_array()[0][0][0], 1.)
        np.testing.assert_array_almost_equal(ds.as_array(), b)

        ds1 = ds.copy()
        ds1 += 1
        ds += ds1
        np.testing.assert_array_almost_equal(ds.as_array(), 3 * b)
        # self.assertEqual(ds.as_array()[0][0][0], 3.)
        ds -= ds1
        np.testing.assert_array_almost_equal(ds.as_array(), b)
        # self.assertEqual(ds.as_array()[0][0][0], 1.)
        ds *= ds1
        np.testing.assert_array_almost_equal(ds.as_array(), 2 * b)
        # self.assertEqual(ds.as_array()[0][0][0], 2.)
        ds /= ds1
        np.testing.assert_array_almost_equal(ds.as_array(), b)
        # self.assertEqual(ds.as_array()[0][0][0], 1.)

    def test_unary_operations(self):
        X, Y, Z = 8, 16, 32
        a = -np.ones((X, Y, Z), dtype='float32')
        b = np.ones((X, Y, Z), dtype='float32')
        ds = DataContainer(a, False, ['X', 'Y', 'Z'])

        ds.sign(out=ds)
        # self.assertEqual(ds.as_array()[0][0][0], -1.)
        np.testing.assert_array_almost_equal(ds.as_array(), -b)


        ds.abs(out=ds)
        np.testing.assert_array_almost_equal(ds.as_array(), b)
        # self.assertEqual(ds.as_array()[0][0][0], 1.)

        ds *= 2
        ds.sqrt(out=ds)
        np.testing.assert_array_almost_equal(ds.as_array(),np.sqrt(2) * b)

        # self.assertEqual(ds.as_array()[0][0][0],
        #                  np.sqrt(2., dtype='float32'))

    def test_binary_add(self):
        X, Y, Z = 8, 16, 32
        a = np.ones((X, Y, Z), dtype='float32')
        b = np.ones((X, Y, Z), dtype='float32')

        ds = DataContainer(a, False, ['X', 'Y', 'Z'])
        ds1 = ds.copy()
        out = ds.copy()

        ds.add(ds1, out=out)
        ds2 = ds.add(ds1)

        self.assertNumpyArrayEqual(out.as_array(), ds2.as_array())

        ds0 = ds
        ds0.add(2, out=out)
        # self.assertEqual(3., out.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(out.as_array(), 3* b)


        ds3 = ds0.add(2)
        self.assertNumpyArrayEqual(out.as_array(), ds3.as_array())

    def test_binary_subtract(self):
        X, Y, Z = 8, 16, 32
        a = np.ones((X, Y, Z), dtype='float32')
        b = np.ones((X, Y, Z), dtype='float32')

        ds = DataContainer(a, False, ['X', 'Y', 'Z'])
        ds1 = ds.copy()
        out = ds.copy()


        ds.subtract(ds1, out=out)
        ds2 = out.subtract(ds1)
        self.assertNumpyArrayEqual(out.as_array(), b * 0)
        # self.assertEqual(0., out.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(ds2.as_array(), -b)
        # self.assertEqual(-1., ds2.as_array()[0][0][0])
        # self.assertLess(t1 - t0, t2 - t1)

        del ds1
        ds0 = ds.copy()
        ds0.subtract(2, out=ds0)
        # self.assertEqual(-1., ds0.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(ds0.as_array(), - b)


        ds3 = ds0.subtract(2)
        # self.assertEqual(-1., ds0.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(ds0.as_array(), - b)

        # self.assertEqual(-3., ds3.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(ds3.as_array(), -3 * b)

    def test_binary_multiply(self):
        X, Y, Z = 8, 16, 32
        a = np.ones((X, Y, Z), dtype='float32')
        b = np.ones((X, Y, Z), dtype='float32')
        ds = DataContainer(a, False, ['X', 'Y', 'Z'])
        ds1 = ds.copy()

        ds.multiply(ds1, out=ds)
        ds2 = ds.multiply(ds1)
        np.testing.assert_array_almost_equal(ds.as_array(), ds2.as_array())

        ds0 = ds
        ds0.multiply(2, out=ds0)
        # self.assertEqual(2., ds0.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(ds0.as_array(), b * 2)

        ds3 = ds0.multiply(2)
        # self.assertEqual(4., ds3.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(ds3.as_array(), 4 * b)

        # self.assertEqual(2., ds.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(ds.as_array(), 2 * b)

        ds.multiply(2.5, out=ds0)
        # self.assertEqual(2.5*2., ds0.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(ds.as_array(), 5 * b)

    def test_binary_divide(self):
        X, Y, Z = 8, 16, 32
        a = np.ones((X, Y, Z), dtype='float32')
        b = np.ones((X, Y, Z), dtype='float32')

        ds = DataContainer(a, False, ['X', 'Y', 'Z'])
        ds1 = ds.copy()

        ds.divide(ds1, out=ds)
        # self.assertEqual(ds.as_array()[0][0][0], 1.)
        np.testing.assert_array_almost_equal(ds.as_array(), b)

        ds0 = ds
        ds0.divide(2, out=ds0)
        # self.assertEqual(0.5, ds0.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(ds0.as_array(), 0.5 * b)


        ds3 = ds0.divide(2)
        # self.assertEqual(.25, ds3.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(ds3.as_array(), 0.25 * b)

        # self.assertEqual(.5, ds.as_array()[0][0][0])
        np.testing.assert_array_almost_equal(ds.as_array(), 0.5 * b)


    def test_reverse_operand_algebra(self):
        number = 3/2

        X, Y, Z = 32, 64, 128
        a = np.ones((X, Y, Z), dtype='float32')
        ds = DataContainer(a * 3, False, ['X', 'Y', 'Z'])

        # rdiv
        b = number / ds
        np.testing.assert_array_almost_equal(a * 0.5, b.as_array())
        # radd
        number = 1
        b = number + ds
        np.testing.assert_array_almost_equal(a * 4, b.as_array())
        # rsub
        number = 3
        b = number - ds
        np.testing.assert_array_almost_equal(np.zeros_like(a), b.as_array())
        # rmul
        number = 1/3
        b = number * ds
        np.testing.assert_array_almost_equal(a, b.as_array())
        # rpow
        number = 2
        b = number ** ds
        np.testing.assert_array_almost_equal(a * 8, b.as_array())

    def test_creation_copy(self):
        shape = 2, 3, 4, 5
        size = np.prod(shape)
        a = np.arange(size)
        #print("a refcount " , sys.getrefcount(a))
        a = np.reshape(a, shape)
        #print("a refcount " , sys.getrefcount(a))
        ds = DataContainer(a, True, ['X', 'Y', 'Z', 'W'])
        #print("a refcount " , sys.getrefcount(a))
        self.assertEqual(sys.getrefcount(a), 2)

    def test_dot(self):
        a0 = np.arange(2*3*4)
        a1 = a0 * 2

        ds0 = DataContainer(np.reshape(a0,(2,3,4)))
        ds1 = DataContainer(np.reshape(a1,(2,3,4)))

        np.testing.assert_equal(ds0.dot(ds1), a0.dot(a1))

        a2 = np.arange(0, 2*3*5*2, 2)
        ds2 = DataContainer(np.reshape(a2,(2,3,5)))

        # it should fail if the shape is wrong
        try:
            ds2.dot(ds0)
            self.assertTrue(False)
        except ValueError as ve:
            self.assertTrue(True)

        n0 = (ds0 * ds1).sum()
        n1 = ds0.as_array().ravel().dot(ds1.as_array().ravel())
        self.assertEqual(n0, n1)

    def test_exp_log(self):
        a0 = np.ones(2*3*4)

        ds0 = DataContainer(np.reshape(a0,(2,3,4)), suppress_warning=True)
        # ds1 = DataContainer(np.reshape(a1,(2,3,4)), suppress_warning=True)
        b = ds0.exp().log()
        np.testing.assert_allclose(ds0.as_array(), b.as_array())

        self.assertEqual(ds0.exp().as_array()[0][0][0], np.exp(1))
        self.assertEqual(ds0.log().as_array()[0][0][0], 0.)

    def test_ImageData(self):
        # create ImageData from geometry
        vgeometry = ImageGeometry(voxel_num_x=4, voxel_num_y=3, channels=2)
        #vol = ImageData(geometry=vgeometry)
        vol = vgeometry.allocate()
        self.assertEqual(vol.shape, (2, 3, 4))

        vol1 = vol + 1
        self.assertNumpyArrayEqual(vol1.as_array(), np.ones(vol.shape))

        vol1 = vol - 1
        self.assertNumpyArrayEqual(vol1.as_array(), -np.ones(vol.shape))

        vol1 = 2 * (vol + 1)
        self.assertNumpyArrayEqual(vol1.as_array(), 2 * np.ones(vol.shape))

        vol1 = (vol + 1) / 2
        self.assertNumpyArrayEqual(vol1.as_array(), np.ones(vol.shape) / 2)

        vol1 = vol + 1
        self.assertEqual(vol1.sum(), 2*3*4)
        vol1 = (vol + 2) ** 2
        self.assertNumpyArrayEqual(vol1.as_array(), np.ones(vol.shape) * 4)

        self.assertEqual(vol.number_of_dimensions, 3)

        ig2 = ImageGeometry (voxel_num_x=2,voxel_num_y=3,voxel_num_z=4,
                     dimension_labels=[ImageDimension["HORIZONTAL_X"], ImageDimension["HORIZONTAL_Y"],
                 ImageDimension["VERTICAL"]])
        data = ig2.allocate()
        self.assertNumpyArrayEqual(np.asarray(data.shape), np.asarray(ig2.shape))
        self.assertNumpyArrayEqual(np.asarray(data.shape), data.as_array().shape)

    def test_ImageData_from_numpy(self):
        """
        Test the creation and manipulation of ImageData from a numpy array.
        """
        # Initialize array and ImageGeometry
        arr = np.arange(24, dtype=np.float32).reshape(2, 3, 4)
        ig = ImageGeometry(voxel_num_x=4, voxel_num_y=3, voxel_num_z=2)
        data = ImageData(arr, deep_copy=True, geometry=ig)

        # Check initial shape, geometry, and deep copy
        self.assertEqual(data.shape, (2, 3, 4))
        self.assertEqual(data.geometry, ig)
        self.assertNotEqual(id(data.array), id(arr))

        # Fill with zeros and check
        data.fill(0)
        np.testing.assert_array_equal(data.as_array(), np.zeros(ig.shape))

        # Fill with original array and check
        data.fill(arr)
        self.assertEqual(data.shape, (2, 3, 4))
        self.assertEqual(data.geometry, ig)

        # Reshape array with singleton dimension and create new ImageData
        arr = arr.reshape(1, 2, 3, 4)
        data = ImageData(arr, geometry=ig)
        self.assertEqual(data.shape, (2, 3, 4))
        self.assertEqual(data.geometry, ig)

        # Fill with zeros and reshaped array, then check
        data.fill(0)
        data.fill(arr)
        self.assertEqual(data.shape, (2, 3, 4))
        self.assertEqual(data.geometry, ig)

    def test_ImageData_apply_circular_mask(self):
        ig = ImageGeometry(voxel_num_x=6, voxel_num_y=3, voxel_size_x=0.5, voxel_size_y = 1)
        data_orig = ig.allocate(1)

        data_masked1 = data_orig.copy()
        data_masked1.apply_circular_mask(0.8)

        self.assertEqual(data_orig.geometry, data_masked1.geometry)
        self.assertEqual(np.count_nonzero(data_masked1.array), 14)

        data_masked1 = data_orig.copy()
        data_masked1.apply_circular_mask(0.5)

        self.assertEqual(data_orig.geometry, data_masked1.geometry)
        self.assertEqual(np.count_nonzero(data_masked1.array), 8)

        data2 = data_orig.copy()
        data_masked2 = data2.apply_circular_mask(0.5, False)

        np.testing.assert_allclose(data_orig.array, data2.array)
        self.assertEqual(np.count_nonzero(data_masked2.array), 8)



    def test_AcquisitionData(self):
        sgeometry = AcquisitionGeometry.create_Parallel3D().set_angles(np.linspace(0, 180, num=10)).set_panel((5,3)).set_channels(2)

        #sino = AcquisitionData(geometry=sgeometry)
        sino = sgeometry.allocate()
        self.assertEqual(sino.shape, (2, 10, 3, 5))

        ag = AcquisitionGeometry.create_Parallel3D().set_angles(np.linspace(0, 180, num=10)).set_panel((2,3)).set_channels(4)
        data = ag.allocate()
        self.assertNumpyArrayEqual(np.asarray(data.shape), np.asarray(ag.shape))
        self.assertNumpyArrayEqual(np.asarray(data.shape), data.as_array().shape)

        ag2 = AcquisitionGeometry.create_Parallel3D().set_angles(np.linspace(0, 180, num=10)).set_panel((2,3)).set_channels(4)\
                                 .set_labels([AcquisitionDimension["VERTICAL"] ,
                         AcquisitionDimension["ANGLE"], AcquisitionDimension["HORIZONTAL"], AcquisitionDimension["CHANNEL"]])

        data = ag2.allocate()
        self.assertNumpyArrayEqual(np.asarray(data.shape), np.asarray(ag2.shape))
        self.assertNumpyArrayEqual(np.asarray(data.shape), data.as_array().shape)

    def test_AcquisitionData_from_numpy(self):
        """
        Test the creation and manipulation of AcquisitionData from a numpy array.
        """
        arr = np.arange(24, dtype=np.float32).reshape(2, 3, 4)
        ag = AcquisitionGeometry.create_Parallel3D().set_angles([0, 1]).set_panel((4, 3))
        data = AcquisitionData(arr, deep_copy=True, geometry=ag)

        # Check initial shape, geometry, and deep copy
        self.assertEqual(data.shape, (2, 3, 4))
        self.assertEqual(data.geometry, ag)
        self.assertNotEqual(id(data.array), id(arr))

        # Fill with zeros and check
        data.fill(0)
        np.testing.assert_array_equal(data.as_array(), np.zeros(ag.shape))

        # Fill with original array and check
        data.fill(arr)
        self.assertEqual(data.shape, (2, 3, 4))
        self.assertEqual(data.geometry, ag)
        np.testing.assert_array_equal(data.as_array(), arr)

        # Reshape array to add empty dimensions and test create AcquisitionData
        arr = arr.reshape(1, 2, 3, 4)
        data = AcquisitionData(arr, geometry=ag)
        self.assertEqual(data.shape, (2, 3, 4))
        self.assertEqual(data.geometry, ag)

        # Fill with reshaped array
        data.fill(0)
        data.fill(arr)
        self.assertEqual(data.shape, (2, 3, 4))
        self.assertEqual(data.geometry, ag)

        # Change dtype to float64, fill, and check
        arr = arr.astype(np.float64)
        data.fill(arr)
        self.assertEqual(data.shape, (2, 3, 4))
        self.assertEqual(data.geometry, ag)
        self.assertEqual(data.dtype, np.float32)


    def test_ImageGeometry_allocate(self):
        vgeometry = ImageGeometry(voxel_num_x=4, voxel_num_y=3, channels=2)
        image = vgeometry.allocate()
        self.assertEqual(0,image.as_array()[0][0][0])
        image = vgeometry.allocate(1)
        self.assertEqual(1,image.as_array()[0][0][0])
        default_order = ['channel' , 'horizontal_y' , 'horizontal_x']
        self.assertEqual(default_order[0], image.dimension_labels[0])
        self.assertEqual(default_order[1], image.dimension_labels[1])
        self.assertEqual(default_order[2], image.dimension_labels[2])
        order = [ 'horizontal_x' , 'horizontal_y', 'channel' ]
        vgeometry.set_labels(order)
        image = vgeometry.allocate(0)
        self.assertEqual(order[0], image.dimension_labels[0])
        self.assertEqual(order[1], image.dimension_labels[1])
        self.assertEqual(order[2], image.dimension_labels[2])

        ig = ImageGeometry(2,3,2)
        try:
            z = ImageData(np.random.randint(10, size=(2,3)), geometry=ig)
            self.assertTrue(False)
        except ValueError as ve:
            log.info(str(ve))
            self.assertTrue(True)

        #vgeometry.allocate('')
    def test_AcquisitionGeometry_allocate(self):
        ageometry = AcquisitionGeometry.create_Parallel3D().set_angles(np.linspace(0, 180, num=10)).set_panel((5,3)).set_channels(2)
        sino = ageometry.allocate(0)
        shape = sino.shape
        self.assertAlmostEqual(0.,sino.as_array()[0][0][0][0])
        self.assertAlmostEqual(0.,sino.as_array()[shape[0]-1][shape[1]-1][shape[2]-1][shape[3]-1])

        sino = ageometry.allocate(1)
        self.assertEqual(1,sino.as_array()[0][0][0][0])
        self.assertEqual(1,sino.as_array()[shape[0]-1][shape[1]-1][shape[2]-1][shape[3]-1])

        default_order = ['channel' , 'angle' ,
                         'vertical' , 'horizontal']
        self.assertEqual(default_order[0], sino.dimension_labels[0])
        self.assertEqual(default_order[1], sino.dimension_labels[1])
        self.assertEqual(default_order[2], sino.dimension_labels[2])
        self.assertEqual(default_order[3], sino.dimension_labels[3])
        order = ['vertical' , 'horizontal', 'channel' , 'angle' ]
        ageometry.set_labels(order)
        sino = ageometry.allocate(0)

        self.assertEqual(order[0], sino.dimension_labels[0])
        self.assertEqual(order[1], sino.dimension_labels[1])
        self.assertEqual(order[2], sino.dimension_labels[2])
        self.assertEqual(order[2], sino.dimension_labels[2])

        try:
            z = AcquisitionData(np.random.randint(10, size=(2,3)), geometry=ageometry)
            self.assertTrue(False)
        except ValueError as ve:
            log.info(str(ve))
            self.assertTrue(True)


    def test_BlockGeometry_allocate_dtype(self):
        ig1 = ImageGeometry(3,3)
        ig2 = ImageGeometry(3,3, dtype=np.int16)
        bg = BlockGeometry(ig1,ig2)

        # print("The default dtype of the BlockImageGeometry is {}".format(bg.dtype))
        self.assertEqual(bg.dtype, (np.float32, np.int16))


    def dtype_allocate_test(self, geometry):
        classname = geometry.__class__.__name__
        # print("The default dtype of the {} is {}".format(classname , geometry.dtype))
        self.assertEqual(geometry.dtype, np.float32)

        #print("Change it to complex")
        geometry.dtype = np.complex64
        self.assertEqual(geometry.dtype, np.complex64)

        geometry.dtype = np.complex128
        self.assertEqual(geometry.dtype, np.complex128)

        geometry.dtype = complex
        self.assertEqual(geometry.dtype, complex)

        #print("Test {} allocate".format(classname ))
        data = geometry.allocate()
        #print("Data dtype is now {} ".format(geometry.dtype))
        self.assertEqual(data.dtype, geometry.dtype)
        #print("Data geometry dtype is now")
        #print(data.geometry.dtype)
        self.assertEqual(data.geometry.dtype, geometry.dtype)

        #print("Allocate data with different dtype, e.g: np.int64 from the same {}".format(classname ))
        data = geometry.allocate(dtype=np.int64)
        self.assertEqual(data.dtype, np.int64)
        #print("Data dtype is now {}".format(data.dtype))
        #print("Data geometry dtype is now {}".format(data.geometry.dtype))
        self.assertEqual(data.geometry.dtype, np.int64)
        self.assertEqual(data.dtype, np.int64)

        #print("The dtype of the {} remain unchanged ig.dtype =  {}".format(classname, geometry.dtype))
        self.assertEqual(geometry.dtype, complex)

        self.assertNotEqual(id(geometry), id(data.geometry))


    def test_ImageGeometry_allocate_dtype(self):
        #print("Test ImageGeometry dtype\n")
        ig = ImageGeometry(3,3)
        self.dtype_allocate_test(ig)


    def test_AcquisitionGeometry_allocate_dtype(self):
        # print("Test AcquisitionGeometry dtype\n")
        # Detectors
        detectors =  10

        # Angles
        angles = np.linspace(0,180,180, dtype='float32')

        # Setup acquisition geometry
        ag = AcquisitionGeometry.create_Parallel2D()\
                                .set_angles(angles)\
                                .set_panel(detectors, pixel_size=0.1)
        self.dtype_allocate_test(ag)


    def test_VectorGeometry_allocate_dtype(self):
        # print("Test VectorGeometry dtype\n")

        vg = VectorGeometry(3)
        self.dtype_allocate_test(vg)


    def complex_allocate_geometry_test(self, geometry):
        data = geometry.allocate(dtype=np.complex64)
        r = (1 + 1j*1)* np.ones(data.shape, dtype=data.dtype)
        data.fill(r)
        self.assertAlmostEqual(data.squared_norm(), data.size * 2)
        np.testing.assert_almost_equal(data.abs().array, np.abs(r))

        data1 = geometry.allocate(dtype=np.float32)
        try:
            data1.fill(r)
            self.assertTrue(False)
        except TypeError as err:
            log.info(str(err))
            self.assertTrue(True)


    def test_ImageGeometry_allocate_complex(self):
        ig = ImageGeometry(2,2)
        self.complex_allocate_geometry_test(ig)

    def test_ImageGeometry_allocate_random_complex(self):
        ig = ImageGeometry(2,2)
        data=ig.allocate('random', dtype=np.complex64)
        self.assertTrue(data.array.dtype, np.complex64)
        self.assertNotEqual(np.sum(data.array).imag, 0)

        ig = ImageGeometry(2,2)
        data=ig.allocate('random_int', dtype=np.complex64)
        self.assertTrue(data.array.dtype, np.complex64)
        self.assertNotEqual(np.sum(data.array).imag, 0)



    def test_AcquisitionGeometry_allocate_complex(self):
        # Detectors
        detectors =  10

        # Angles
        angles = np.linspace(0,10,10, dtype='float32')

        # Setup acquisition geometry
        ag = AcquisitionGeometry.create_Parallel2D()\
                                .set_angles(angles)\
                                .set_panel(detectors, pixel_size=0.1)

        self.complex_allocate_geometry_test(ag)

    def test_AcquisitionGeometry_allocate_random_complex(self):

        detectors =  10

        # Angles
        angles = np.linspace(0,10,10, dtype='float32')

        # Setup acquisition geometry
        ag = AcquisitionGeometry.create_Parallel2D()\
                                .set_angles(angles)\
                                .set_panel(detectors, pixel_size=0.1)


        data=ag.allocate('random', dtype=np.complex64)
        self.assertTrue(data.array.dtype, np.complex64)
        self.assertNotEqual(np.sum(data.array).imag, 0)


        data=ag.allocate('random_int', dtype=np.complex64)
        self.assertTrue(data.array.dtype, np.complex64)
        self.assertNotEqual(np.sum(data.array).imag, 0)



    def test_VectorGeometry_allocate_complex(self):
        vg = VectorGeometry(3)
        self.complex_allocate_geometry_test(vg)

        data=vg.allocate('random', dtype=np.complex64)
        self.assertTrue(data.array.dtype, np.complex64)
        self.assertNotEqual(np.sum(data.array).imag, 0)

        data=vg.allocate('random_int', dtype=np.complex64)
        self.assertTrue(data.array.dtype, np.complex64)
        self.assertNotEqual(np.sum(data.array).imag, 0)


    def test_ImageGeometry_allocate_random_same_seed(self):
        vgeometry = ImageGeometry(voxel_num_x=4, voxel_num_y=3, channels=2)
        image1 = vgeometry.allocate('random', seed=0)
        image2 = vgeometry.allocate('random', seed=0)
        np.testing.assert_allclose(image1.as_array(), image2.as_array())


    def test_AcquisitionDataSubset(self):
        sgeometry = AcquisitionGeometry.create_Parallel3D().set_angles(np.linspace(0, 180, num=10)).set_panel((5,3)).set_channels(2)

        # expected dimension_labels

        self.assertListEqual([AcquisitionDimension["CHANNEL"] ,
                 AcquisitionDimension["ANGLE"] , AcquisitionDimension["VERTICAL"] ,
                 AcquisitionDimension["HORIZONTAL"]],
                              list(sgeometry.dimension_labels))
        sino = sgeometry.allocate()

        # test reshape
        new_order = [AcquisitionDimension["HORIZONTAL"] ,
                 AcquisitionDimension["CHANNEL"] , AcquisitionDimension["VERTICAL"] ,
                 AcquisitionDimension["ANGLE"]]
        sino.reorder(new_order)

        self.assertListEqual(new_order, list(sino.geometry.dimension_labels))

        ss1 = sino.get_slice(vertical = 0)
        self.assertListEqual([AcquisitionDimension["HORIZONTAL"] ,
                 AcquisitionDimension["CHANNEL"]  ,
                 AcquisitionDimension["ANGLE"]], list(ss1.geometry.dimension_labels))
        ss2 = sino.get_slice(vertical = 0, channel=0)
        self.assertListEqual([AcquisitionDimension["HORIZONTAL"] ,
                 AcquisitionDimension["ANGLE"]], list(ss2.geometry.dimension_labels))


    def test_ImageDataSubset(self):
        new_order = ['horizontal_x', 'channel', 'horizontal_y']


        vgeometry = ImageGeometry(voxel_num_x=4, voxel_num_y=3, channels=2, dimension_labels=new_order)
        # expected dimension_labels

        self.assertListEqual(new_order,
                              list(vgeometry.dimension_labels))
        vol = vgeometry.allocate()

        # test reshape
        new_order = ['channel', 'horizontal_y','horizontal_x']
        vol.reorder(new_order)

        self.assertListEqual(new_order, list(vol.geometry.dimension_labels))

        ss1 = vol.get_slice(horizontal_x = 0)
        self.assertListEqual(['channel', 'horizontal_y'], list(ss1.geometry.dimension_labels))

        vg = ImageGeometry(3,4,5,channels=2)
        self.assertListEqual([ImageDimension["CHANNEL"], ImageDimension["VERTICAL"],
                ImageDimension["HORIZONTAL_Y"], ImageDimension["HORIZONTAL_X"]],
                              list(vg.dimension_labels))
        ss2 = vg.allocate()
        ss3 = vol.get_slice(channel=0)
        self.assertListEqual([ImageDimension["HORIZONTAL_Y"], ImageDimension["HORIZONTAL_X"]], list(ss3.geometry.dimension_labels))

    def test_DataContainerSubset(self):
        dc = DataContainer(np.ones((2,3,4,5)))

        dc.dimension_labels =[AcquisitionDimension["CHANNEL"] ,
                 AcquisitionDimension["ANGLE"] , AcquisitionDimension["VERTICAL"] ,
                 AcquisitionDimension["HORIZONTAL"]]

        # test reshape
        new_order = [AcquisitionDimension["HORIZONTAL"] ,
                 AcquisitionDimension["CHANNEL"] , AcquisitionDimension["VERTICAL"] ,
                 AcquisitionDimension["ANGLE"]]
        dc.reorder(new_order)

        self.assertListEqual(new_order, list(dc.dimension_labels))

        ss1 = dc.get_slice(vertical=0)

        self.assertListEqual([AcquisitionDimension["HORIZONTAL"] ,
                 AcquisitionDimension["CHANNEL"]  ,
                 AcquisitionDimension["ANGLE"]], list(ss1.dimension_labels))

        ss2 = dc.get_slice(vertical=0, channel=0)
        self.assertListEqual([AcquisitionDimension["HORIZONTAL"] ,
                 AcquisitionDimension["ANGLE"]], list(ss2.dimension_labels))

        # Check we can get slice still even if force parameter is passed:
        ss3 = dc.get_slice(vertical=0, channel=0, force=True)
        self.assertListEqual([AcquisitionDimension["HORIZONTAL"] ,
                    AcquisitionDimension["ANGLE"]], list(ss3.dimension_labels))


    def test_DataContainerChaining(self):
        dc = self.create_DataContainer(256,256,256,1)

        dc.add(9,out=dc)\
          .subtract(1,out=dc)
        self.assertEqual(1+9-1,dc.as_array().flatten()[0])


    def test_reduction(self):
        dc = self.create_DataContainer(2,2,2,value=1)
        sqnorm = dc.squared_norm()
        norm = dc.norm()
        self.assertEqual(sqnorm, 8.0)
        np.testing.assert_almost_equal(norm, np.sqrt(8.0), decimal=7)
        sum = dc.sum(axis=('X','Z'))
        np.testing.assert_almost_equal(sum.as_array(), [np.float64(4),np.float64(4)])
        np.testing.assert_equal(sum.dimension_labels,('Y',))
        sum = dc.sum()
        np.testing.assert_almost_equal(sum, 8.0)


    def test_reduction_mean(self):
        ig = ImageGeometry(2,2)
        data = ig.allocate(0)
        np_arr = data.as_array()
        np_arr[0][0] = 0
        np_arr[0][1] = 1
        np_arr[1][0] = 2
        np_arr[1][1] = 3
        data.fill(np_arr)

        mean = data.mean()
        expected = np.float64(0+1+2+3)/np.float64(4)
        np.testing.assert_almost_equal(mean, expected)


    def directional_reduction_unary_test(self, data, test_func, expected_func, out, function_name):
        def error_message(function_name, test_name):
            return "Failed with reduction " + function_name + " on test " + test_name
        # test specifying function in 1 axis
        result = test_func(axis=data.dimension_labels[1])
        expected = expected_func(data.as_array(), axis=1)
        expected_dimension_labels = data.dimension_labels[0],data.dimension_labels[2]
        np.testing.assert_almost_equal(result.as_array(), expected, err_msg=error_message(function_name, "'with 1 axis'"))
        self.assertEqual(result.dimension_labels, expected_dimension_labels, f"{function_name} 'with 1 axis'")
        # test specifying axis with an int
        result = test_func(axis=1)
        np.testing.assert_almost_equal(result.as_array(), expected, err_msg=error_message(function_name, "'with 1 axis'"))
        self.assertEqual(result.dimension_labels,expected_dimension_labels, f"{function_name} 'with 1 axis'")
        # test specifying function in 2 axes
        result = test_func(axis=(data.dimension_labels[0],data.dimension_labels[1]))
        np.testing.assert_almost_equal(result.as_array(), expected_func(data.as_array(), axis=(0,1)), err_msg=error_message(function_name, "'with 2 axes'"))
        self.assertEqual(result.dimension_labels, (data.dimension_labels[2],), f"{function_name} 'with 2 axes'")
        # test specifying function in 2 axes with an int
        result = test_func(axis=(0,1))
        np.testing.assert_almost_equal(result.as_array(), expected_func(data.as_array(), axis=(0,1)), err_msg=error_message(function_name, "'with 2 axes'"))
        self.assertEqual(result.dimension_labels, (data.dimension_labels[2],), f"{function_name} 'with 2 axes'")
        # test specifying function in 3 axes
        result = test_func(axis=(data.dimension_labels[0],data.dimension_labels[1],data.dimension_labels[2]))
        np.testing.assert_almost_equal(result, expected_func(data.as_array()), err_msg=error_message(function_name, "'with 3 axes'"))
        # test providing a DataContainer to out
        expected_array = expected_func(data.as_array(), axis = 0)
        test_func(axis=0, out=out)
        np.testing.assert_almost_equal(out.as_array(), expected_array, err_msg=error_message(function_name, "'of out argument'"))
        self.assertEqual(out.dimension_labels, (data.dimension_labels[1],data.dimension_labels[2]), f"{function_name} 'of out argument'")
        test_func(axis=data.dimension_labels[0], out=out)
        np.testing.assert_almost_equal(out.as_array(), expected_array, err_msg=error_message(function_name, "'of out argument'"))
        self.assertEqual(out.dimension_labels, (data.dimension_labels[1],data.dimension_labels[2]), f"{function_name} 'of out argument'")
        # test providing a numpy array to out
        out = np.zeros((2,2), dtype=data.dtype)
        test_func(axis=0, out=out)
        np.testing.assert_almost_equal(out, expected_array, err_msg=error_message(function_name, "'of out argument'"))

    def test_directional_reduction_unary(self):
        np_arr = np.array([[[0,1],[2,3]],[[4,5],[6,7]]], dtype=np.float32)
        # create DataContainer test class
        dc =  DataContainer(np_arr, dimension_labels=('vertical', 'horizontal_y', 'horizontal_x'))
        dc_out = DataContainer(np.zeros((2,2)),dimension_labels=('horizontal_y', 'horizontal_x'))
        # create ImageData test class
        id = ImageGeometry(2,2,2).allocate(0)
        id.fill(np_arr)
        id_out = ImageGeometry(2,2).allocate(0)
        # create complex ImageData test class
        id_complex = ImageGeometry(2,2,2).allocate(0, dtype=complex)
        complex_arr = np.empty((2,2,2), dtype=complex)
        complex_arr.real = np_arr
        complex_arr.imag = np.array([[[7,6],[5,4]],[[3,2],[1,0]]])
        id_complex.fill(complex_arr)
        id_complex_out = ImageGeometry(2,2).allocate(0, dtype=complex)
        id_complex_out.fill(np.zeros((2,2), dtype=complex))
        # create AcquisitionData test class
        ag = AcquisitionGeometry.create_Parallel3D().set_angles(np.linspace(0, 180, num=2)).set_panel((2,2))
        ad = ag.allocate()
        ad.fill(np_arr)
        ag = AcquisitionGeometry.create_Parallel3D().set_angles(np.linspace(0, 180, num=0)).set_panel((2,2))
        ad_out = ag.allocate()
        ad_out.fill(np.zeros((2,2)))

        data_classes = [dc, id, id_complex, ad]
        out_classes = [dc_out,id_out,id_complex_out, ad_out]
        for j in np.arange(len(data_classes)):
            function_names = ['mean','sum','min','max']
            for i in np.arange(len(function_names)):
                self.directional_reduction_unary_test(data_classes[j], getattr(data_classes[j],function_names[i]), getattr(np, function_names[i]), out_classes[j], function_names[i])


    def test_mean_direction(self):
        np_arr = np.array([[[0,1],[2,3]],[[4,5],[6,7]]])
        ig = ImageGeometry(2,2,2)
        data = ig.allocate(0)
        data.fill(np_arr)

        # test specifying mean in 1 axis
        mean = data.mean(axis='horizontal_y')
        np.testing.assert_almost_equal(mean.as_array(), [[1.0, 2.0],[5.0, 6.0]])
        np.testing.assert_equal(mean.dimension_labels,('vertical','horizontal_x'))
        mean = data.mean(axis=1)
        np.testing.assert_almost_equal(mean.as_array(), [[1.0, 2.0],[5.0, 6.0]])
        # test specifying mean in 2 axes
        mean = data.mean(axis=('horizontal_y', 'vertical'))
        np.testing.assert_almost_equal(mean.as_array(), [3.0, 4.0])
        np.testing.assert_equal(mean.dimension_labels,('horizontal_x',))
        # test specifying mean in 3 axes
        expected = np.float64(0+1+2+3+4+5+6+7)/np.float64(8)
        mean = data.mean(axis=('horizontal_x','horizontal_y','vertical'))
        np.testing.assert_almost_equal(mean, expected)
        # test mean on VectorData
        np_arr = np.array([0,1,2,3,4])
        vg = VectorGeometry(5)
        vd = vg.allocate(0)
        vd.fill(np_arr)
        vd.dimension_labels = 'x'
        np.testing.assert_almost_equal(vd.mean(axis='x'), np.mean(vd))


    def test_multiply_out(self):
        ig = ImageGeometry(10,11,12)
        u = ig.allocate()
        a = np.ones(u.shape)

        u.fill(a)

        np.testing.assert_array_equal(a, u.as_array())

        #u = ig.allocate(ImageDimension["RANDOM_INT"], seed=1)
        l = functools.reduce(lambda x,y: x*y, (10,11,12), 1)

        a = np.zeros((l, ), dtype=np.float32)
        for i in range(l):
            a[i] = np.sin(2 * i* 3.1415/l)
        b = np.reshape(a, u.shape)
        u.fill(b)
        np.testing.assert_array_equal(b, u.as_array())

        u.multiply(2, out=u)
        c = b * 2
        np.testing.assert_allclose(u.as_array(), c)


    def test_sapyb_datacontainer_f(self):
        #a vec, b vec

        ig = ImageGeometry(10,10)
        d1 = ig.allocate(dtype=np.float32)
        d2 = ig.allocate(dtype=np.float32)
        a = ig.allocate(dtype=np.float32)
        b = ig.allocate(dtype=np.float32)

        d1.fill(np.asarray(np.arange(1,101).reshape(10,10), dtype=np.float32))
        d2.fill(np.asarray(np.arange(1,101).reshape(10,10), dtype=np.float32))
        a.fill(1.0/d1.as_array())
        b.fill(-1.0/d2.as_array())

        out = ig.allocate(-1,dtype=np.float32)
        # equals to 1 + -1 = 0
        out = d1.sapyb(a,d2,b)
        res = np.zeros_like(d1.as_array())
        np.testing.assert_array_equal(res, out.as_array())

        out.fill(0)
        d1.sapyb(a,d2,b, out)
        res = np.zeros_like(d1.as_array())
        np.testing.assert_array_equal(res, out.as_array())


    def test_sapyb_scalar_f(self):
        # a,b scalar
        ig = ImageGeometry(10,10)
        d1 = ig.allocate(1, dtype=np.float32)
        d2 = ig.allocate(2, dtype=np.float32)
        a = 2.
        b = -1.

        out = ig.allocate(-1,dtype=np.float32)
        # equals to 2*[1] + -1*[2] = 0
        out = d1.sapyb(a,d2,b)
        res = np.zeros_like(d1.as_array())
        np.testing.assert_array_equal(res, out.as_array())

        out.fill(0)
        d1.sapyb(a,d2,b, out)
        np.testing.assert_array_equal(res, out.as_array())

        d1.sapyb(a,d2,b, out=d1)
        np.testing.assert_array_equal(res, d1.as_array())

        d1.fill(1)
        d1.sapyb(a,d2,b, out=d2)
        np.testing.assert_array_equal(res, d2.as_array())


    def test_sapyb_datacontainer_scalar_f(self):
        #mix: a scalar and b DataContainer and a DataContainer and b scalar
        ig = ImageGeometry(10,10)
        d1 = ig.allocate(1., dtype=np.complex64)
        d2 = ig.allocate(2.,dtype=np.complex64)
        a = 2.+2j
        b = ig.allocate(-1.-1j, dtype=np.complex64)

        out = ig.allocate(-1,dtype=np.complex64)
        # equals to (2+2j)*[1] + -(1+j)*[2] = 0

        out = d1.sapyb(a,d2,b)
        res = ig.allocate(0, dtype=np.complex64)
        np.testing.assert_array_equal(res.as_array(), out.as_array())

        out.fill(-1)
        d1.sapyb(a,d2,b, out)
        np.testing.assert_array_equal(res.as_array(), out.as_array())

        out = d2.sapyb(b,d1,a)
        res = ig.allocate(0, dtype=np.complex64)
        np.testing.assert_array_equal(res.as_array(), out.as_array())

        out.fill(-1)
        d2.sapyb(b,d1,a, out)
        np.testing.assert_array_equal(res.as_array(), out.as_array())


    def test_sapyb_scalar_c(self):
        # a, b scalar
        ig = ImageGeometry(10,10)
        d1 = ig.allocate(1, dtype=np.complex64)
        d2 = ig.allocate(2, dtype=np.complex64)
        a = 2.+2j
        b = -1.-1j

        out = ig.allocate(-1,dtype=np.complex64)
        # equals to (2+2j)*[1] + -(1+j)*[2] = 0
        out = d1.sapyb(a,d2,b)
        res = np.zeros_like(d1.as_array())
        np.testing.assert_array_equal(res, out.as_array())

        out.fill(0)
        d1.sapyb(a,d2,b, out)
        np.testing.assert_array_equal(res, out.as_array())

        d1.sapyb(a,d2,b, out=d1)
        np.testing.assert_array_equal(res, d1.as_array())

        d1.fill(1)
        d1.sapyb(a,d2,b, out=d2)
        np.testing.assert_array_equal(res, d2.as_array())


    def test_sapyb_datacontainer_c(self):
        #a vec, b vec
        ig = ImageGeometry(10,10)
        d1 = ig.allocate(dtype=np.complex64)
        d2 = ig.allocate(dtype=np.complex64)
        a = ig.allocate(dtype=np.complex64)
        b = ig.allocate(dtype=np.complex64)

        arr = np.empty(ig.shape, dtype=np.complex64)
        arr.real = np.asarray(np.arange(1,101).reshape(10,10), dtype=np.float32)
        arr.imag = np.asarray(np.arange(1,101).reshape(10,10), dtype=np.float32)

        d1.fill(arr)

        arr.imag = -1* arr.imag
        d2.fill(arr)

        a.fill(d2.as_array())
        b.fill(d1.as_array())

        out = ig.allocate(-1,dtype=np.complex64)
        # equals to d1^ * d1 + d2^*d2 = d1**2 + d2**2 = 2* arr.norm = 2 * (arr.real **2 + arr.imag **2)
        out = d1.sapyb(a,d2,b)
        res = 2* (arr.real * arr.real + arr.imag * arr.imag)
        np.testing.assert_array_equal(res, out.as_array())

        out.fill(0)
        d1.sapyb(a,d2,b, out)
        np.testing.assert_array_equal(res, out.as_array())


    def test_sapyb_datacontainer_scalar_c(self):
        #mix: a scalar and b DataContainer and a DataContainer and b scalar
        ig = ImageGeometry(10,10)
        d1 = ig.allocate(1., dtype=np.complex64)
        d2 = ig.allocate(2.,dtype=np.complex64)
        a = 2.+2j
        b = ig.allocate(-1.-1j, dtype=np.complex64)


        out = ig.allocate(-1,dtype=np.complex64)
        # equals to (2+2j)*[1] + -(1+j)*[2] = 0

        out = d1.sapyb(a,d2,b)
        res = ig.allocate(0, dtype=np.complex64)
        np.testing.assert_array_equal(res.as_array(), out.as_array())

        out.fill(-1)
        d1.sapyb(a,d2,b, out)
        np.testing.assert_array_equal(res.as_array(), out.as_array())

        out = d2.sapyb(b,d1,a)
        res = ig.allocate(0, dtype=np.complex64)
        np.testing.assert_array_equal(res.as_array(), out.as_array())

        out.fill(-1)
        d2.sapyb(b,d1,a, out)
        np.testing.assert_array_equal(res.as_array(), out.as_array())


    def test_sapyb_scalar_f_c(self):
        # a,b scalar
        ig = ImageGeometry(10,10)
        d1 = ig.allocate(1, dtype=np.complex64)
        d2 = ig.allocate(2, dtype=np.float32)
        a = 2.+1j
        b = -1.

        # equals to 2*[1] + -1*[2] = 0
        out = d1.sapyb(a,d2,b)
        res = np.zeros_like(d1.as_array()) + 1j
        np.testing.assert_array_equal(res, out.as_array())

        out.fill(0)
        d1.sapyb(a,d2,b, out)
        np.testing.assert_array_equal(res, out.as_array())

        d1.sapyb(a,d2,b, out=d1)
        np.testing.assert_array_equal(res, d1.as_array())

        d1.fill(1)
        try:
            with self.assertRaises(np.core._exceptions.UFuncTypeError) as context:
                d1.sapyb(a,d2,b, out=d2)
        except AttributeError as ae:
            log.info("Probably numpy version too low: %s", ae)

        # print ("Exception thrown:", str(context.exception))

        # out is complex
        # d1.fill(1+0j)
        d2.fill(2)
        d1.sapyb(a,d2,b,out=d1)
        # 2+1j * [1+0j] -1 * [2]
        np.testing.assert_array_equal(1j * np.ones_like(d1.as_array()), d1.as_array())


    def test_min(self):
        ig = ImageGeometry(10,10)
        a = np.asarray(np.linspace(-10,10, num=100, endpoint=True), dtype=np.float32)
        a = a.reshape((10,10))
        d1 = ig.allocate(1)
        d1.fill(a)
        self.assertAlmostEqual(d1.min(), -10.)


    def test_min_direction(self):
        ig = ImageGeometry(2,2,2)
        data = ig.allocate(0)
        np_arr = np.array([[[0,1],[2,3]],[[4,5],[6,7]]])
        data.fill(np_arr)

        # test specifying min in 3 axes
        min = data.min(axis=('vertical','horizontal_x','horizontal_y'))
        self.assertAlmostEqual(min, 0)
        # test specifying min in 2 axes
        min = data.min(axis=('horizontal_y', 'vertical'))
        np.testing.assert_almost_equal(min.as_array(), [0.0, 1.0])
        # test specifying min in 1 axis
        min = data.min(axis='horizontal_x')
        expected = [[np.float64(0), np.float64(2)],[np.float64(4), np.float64(6)]]
        np.testing.assert_almost_equal(min.as_array(), expected)
        np.testing.assert_equal(min.dimension_labels,('vertical','horizontal_y'))
        # test specifying min in 1 axis using numpy axis argument
        min = data.min(axis=0)
        expected = [[np.float64(0), np.float64(1)],[np.float64(2), np.float64(3)]]
        np.testing.assert_almost_equal(min.as_array(), expected)


    def test_max(self):
        ig = ImageGeometry(10,10)
        a = np.asarray(np.linspace(-10,10, num=100, endpoint=True), dtype=np.float32)
        a = a.reshape((10,10))
        d1 = ig.allocate(1)
        d1.fill(a)
        self.assertAlmostEqual(d1.max(), 10.)


    def test_max_direction(self):
        ig = ImageGeometry(2,2,2)
        data = ig.allocate(0)
        np_arr = np.array([[[0,1],[2,3]],[[4,5],[6,7]]])
        data.fill(np_arr)

        # test specifying max in 3 axes
        max = data.max(axis=('vertical','horizontal_x','horizontal_y'))
        self.assertAlmostEqual(max, 7)
        # test specifying max in 2 axes
        max = data.max(axis=('horizontal_y', 'vertical'))
        np.testing.assert_almost_equal(max.as_array(), [6.0, 7.0])
        np.testing.assert_equal(max.dimension_labels,('horizontal_x',))
        # test specifying max in 1 axis
        max = data.max(axis='horizontal_x')
        expected = [[np.float64(1), np.float64(3)],[np.float64(5), np.float64(7)]]
        np.testing.assert_almost_equal(max.as_array(), expected)
        np.testing.assert_equal(max.dimension_labels,('vertical','horizontal_y'))
        # test specifying max in 1 axis using numpy axis argument
        max = data.max(axis=0)
        expected = [[np.float64(4), np.float64(5)],[np.float64(6), np.float64(7)]]
        np.testing.assert_almost_equal(max.as_array(), expected)


    def test_size(self):
        ig = ImageGeometry(10,10)
        d1 = ig.allocate(1)

        self.assertEqual( d1.size, 100 )

        sgeometry = AcquisitionGeometry.create_Parallel3D().set_angles(np.linspace(0, 180, num=10)).set_panel((5,3)).set_channels(2)

        ad = sgeometry.allocate()

        self.assertEqual( ad.size, 3*5*10*2 )


    def test_negation(self):
        X, Y, Z = 256, 512, 512
        a = np.ones((X, Y, Z), dtype='int32')

        ds = - DataContainer(a, False, ['X', 'Y', 'Z'])

        np.testing.assert_array_equal(ds.as_array(), -a)


    def test_fill_dimension_ImageData(self):
        ig = ImageGeometry(2,3,4)
        u = ig.allocate(0)
        a = np.ones((4,2))
        # default_labels = [ImageDimension["VERTICAL"], ImageDimension["HORIZONTAL_Y"], ImageDimension["HORIZONTAL_X"]]

        data = u.as_array()
        axis_number = u.get_dimension_axis('horizontal_y')

        u.fill(a, horizontal_y=0)
        np.testing.assert_array_equal(u.get_slice(horizontal_y=0).as_array(), a)

        u.fill(2, horizontal_y=1)
        np.testing.assert_array_equal(u.get_slice(horizontal_y=1).as_array(), 2 * a)

        u.fill(2, horizontal_y=1)
        np.testing.assert_array_equal(u.get_slice(horizontal_y=1).as_array(), 2 * a)

        b = u.get_slice(horizontal_y=2)
        b.fill(3)
        u.fill(b, horizontal_y=2)
        np.testing.assert_array_equal(u.get_slice(horizontal_y=2).as_array(), 3 * a)


    def test_fill_dimension_AcquisitionData(self):
        ag = AcquisitionGeometry.create_Parallel3D()
        ag.set_channels(4)
        ag.set_panel([2,3])
        ag.set_angles([0,1,2,3,5])
        ag.set_labels(('horizontal','angle','vertical','channel'))
        u = ag.allocate(0)
        a = np.ones((4,2))
        # default_labels = [ImageDimension["VERTICAL"], ImageDimension["HORIZONTAL_Y"], ImageDimension["HORIZONTAL_X"]]

        data = u.as_array()
        axis_number = u.get_dimension_axis('horizontal_y')

        u.fill(a, horizontal_y=0)
        np.testing.assert_array_equal(u.subset(horizontal_y=0).as_array(), a)

        u.fill(2, horizontal_y=1)
        np.testing.assert_array_equal(u.subset(horizontal_y=0).as_array(), 2 * a)

        u.fill(2, horizontal_y=1)
        np.testing.assert_array_equal(u.subset(horizontal_y=1).as_array(), 2 * a)

        b = u.subset(horizontal_y=2)
        b.fill(3)
        u.fill(b, horizontal_y=2)
        np.testing.assert_array_equal(u.subset(horizontal_y=2).as_array(), 3 * a)

        # slice with 2 axis
        a = np.ones((2,))
        u.fill(a, horizontal_y=1, vertical=0)
        np.testing.assert_array_equal(u.subset(horizontal_y=1, vertical=0).as_array(), a)


    def test_fill_dimension_AcquisitionData(self):
        ag = AcquisitionGeometry.create_Parallel3D()
        ag.set_channels(4)
        ag.set_panel([2,3])
        ag.set_angles([0,1,2,3,5])
        ag.set_labels(('horizontal','angle','vertical','channel'))
        u = ag.allocate(0)
        # (2, 5, 3, 4)
        a = np.ones((2,5))
        # default_labels = [ImageDimension["VERTICAL"], ImageDimension["HORIZONTAL_Y"], ImageDimension["HORIZONTAL_X"]]
        b = u.get_slice(channel=0, vertical=0)
        data = u.as_array()

        u.fill(a, channel=0, vertical=0)
        np.testing.assert_array_equal(u.get_slice(channel=0, vertical=0).as_array(), a)

        u.fill(2, channel=0, vertical=0)
        np.testing.assert_array_equal(u.get_slice(channel=0, vertical=0).as_array(), 2 * a)

        u.fill(2, channel=0, vertical=0)
        np.testing.assert_array_equal(u.get_slice(channel=0, vertical=0).as_array(), 2 * a)

        b = u.get_slice(channel=0, vertical=0)
        b.fill(3)
        u.fill(b, channel=1, vertical=1)
        np.testing.assert_array_equal(u.get_slice(channel=1, vertical=1).as_array(), 3 * a)


    def test_vectordata_dot_product(self):
        x = np.array([1 + 1j, 2 - 5j])
        y = np.array([3 - 2j, 1 + 1j])

        a1 = np.dot(y, x.conjugate())
        a2 = np.dot(x.conjugate(), y)
        self.assertAlmostEqual( a1, a2)
        b1 = np.vdot(x, y)
        b2 = np.vdot(y, x).conjugate()
        self.assertAlmostEqual( b1, b2)

        print(a1, b1)
        print(a2, b2)

        xcil = VectorData(x)
        ycil = VectorData(y)
        self.assertAlmostEqual( ycil.dot(xcil), xcil.dot(ycil).conjugate())



#  Copyright 2022 United Kingdom Research and Innovation
#  Copyright 2022 The University of Manchester
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
import unittest
from cil.optimisation.functions import KullbackLeibler
from cil.framework import ImageGeometry
import numpy
import scipy
from utils import has_numba, initialise_tests

initialise_tests()

if has_numba:
    import numba

class TestKullbackLeiblerNumpy(unittest.TestCase):

    def setUp(self):

        M, N, K =  2, 3, 4
        self.ig = ImageGeometry(N, M, K)

        self.u1 = self.ig.allocate('random', seed = 500)
        self.g1 = self.ig.allocate('random', seed = 100)
        self.b1 = self.ig.allocate('random', seed = 1000)

        self.f = KullbackLeibler(b = self.g1, backend='numpy')
        self.f1 = KullbackLeibler(b = self.g1, eta = self.b1,  backend='numpy')
        self.tau = 400.4

    def test_signature(self):

        # with no data
        with self.assertRaises(TypeError):
            f = KullbackLeibler()

        with self.assertRaises(ValueError):
            f = KullbackLeibler(b=-1*self.g1)

    def test_call_method(self):

        # without eta
        numpy.testing.assert_allclose(0.0, self.f(self.g1))

        # with eta
        tmp_sum = (self.u1 + self.f1.eta).as_array()
        ind = tmp_sum >= 0
        tmp = scipy.special.kl_div(self.f1.b.as_array()[ind], tmp_sum[ind])
        numpy.testing.assert_allclose(self.f1(self.u1), numpy.sum(tmp) )

    def test_gradient_method(self):

        # without eta
        res1 = self.f.gradient(self.u1)
        res2 = self.u1.geometry.allocate()
        self.f.gradient(self.u1, out = res2)
        numpy.testing.assert_allclose(res1.as_array(), res2.as_array())

        # with eta
        res1 = self.f1.gradient(self.u1)
        res2 = self.u1.geometry.allocate()
        self.f1.gradient(self.u1, out = res2)
        numpy.testing.assert_allclose(res1.as_array(), res2.as_array())

    def test_proximal_method(self):

        # without eta
        res1 = self.f.proximal(self.u1, self.tau)
        res2 = self.u1.geometry.allocate()
        self.f.proximal(self.u1, self.tau, out = res2)
        numpy.testing.assert_array_almost_equal(res1.as_array(), res2.as_array(), decimal=4)

        # with eta
        res1 = self.f1.proximal(self.u1, self.tau)
        res2 = self.u1.geometry.allocate()
        self.f1.proximal(self.u1, self.tau, out = res2)
        numpy.testing.assert_array_almost_equal(res1.as_array(), res2.as_array(), decimal=4)

    def test_proximal_conjugate_method(self):

        # without eta
        res1 = self.f.proximal_conjugate(self.u1, self.tau)
        res2 = self.u1.geometry.allocate()
        self.f.proximal_conjugate(self.u1, self.tau, out = res2)
        numpy.testing.assert_array_almost_equal(res1.as_array(), res2.as_array(), decimal=4)

        # with eta
        res1 = self.f1.proximal_conjugate(self.u1, self.tau)
        res2 = self.u1.geometry.allocate()
        self.f1.proximal_conjugate(self.u1, self.tau, out = res2)
        numpy.testing.assert_array_almost_equal(res1.as_array(), res2.as_array(), decimal=4)

    def test_convex_conjugate_method(self):

        # with eta
        tmp = 1 - self.u1.as_array()
        ind = tmp>0
        xlogy = - scipy.special.xlogy(self.f1.b.as_array()[ind], tmp[ind])
        res1 = numpy.sum(xlogy) - self.f1.eta.dot(self.u1)
        res2 = self.f1.convex_conjugate(self.u1)
        numpy.testing.assert_equal(res1, res2)

        # without eta
        tmp = 1 - self.u1.as_array()
        ind = tmp>0
        xlogy = - scipy.special.xlogy(self.f.b.as_array()[ind], tmp[ind])
        res1 = numpy.sum(xlogy) - self.f.eta.dot(self.u1)
        res2 = self.f.convex_conjugate(self.u1)
        numpy.testing.assert_equal(res1, res2)

@unittest.skipUnless(has_numba, "Skipping because numba isn't installed")
class TestKullbackLeiblerNumba(unittest.TestCase):
    def setUp(self):
        #numpy.random.seed(1)
        M, N, K =  2, 3, 4
        ig = ImageGeometry(N, M)

        u1 = ig.allocate('random', seed = 500)
        u1 = ig.allocate(0.2)
        #g1 = ig.allocate('random', seed = 100)
        g1 = ig.allocate(1)

        b1 = ig.allocate('random', seed = 1000)
        eta = ig.allocate(1e-3)

        mask = ig.allocate(1)

        mask.fill(0, horizontal_x=0)

        mask_c = ig.allocate(0)
        mask_c.fill(1, horizontal_x=0)

        f = KullbackLeibler(b=g1, backend="numba", eta=eta)
        f_np = KullbackLeibler(b=g1, backend="numpy", eta=eta)

        # mask is on vartical=0
        # separate the u1 vertical=0
        f_mask = KullbackLeibler(b=g1.copy(), backend="numba", mask=mask.copy(), eta=eta.copy())
        f_mask_c = KullbackLeibler(b=g1.copy(), backend="numba", mask=mask_c.copy(), eta=eta.copy())
        f_on_mask = KullbackLeibler(b=g1.get_slice(horizontal_x=0), backend="numba", eta=eta.get_slice(horizontal_x=0))
        u1_on_mask = u1.get_slice(horizontal_x=0)

        tau = 400.4
        self.tau = tau
        self.u1 = u1
        self.g1 = g1
        self.b1 = b1
        self.eta = eta
        self.f = f
        self.f_np = f_np
        self.mask = mask
        self.mask_c = mask_c
        self.f_mask = f_mask
        self.f_mask_c = f_mask_c
        self.f_on_mask = f_on_mask
        self.u1_on_mask = u1_on_mask

    def test_KullbackLeibler_numba_call(self):
        f = self.f
        f_np = self.f_np
        tau = self.tau
        u1 = self.u1

        numpy.testing.assert_allclose(f(u1), f_np(u1),  rtol=1e-5)

    def test_KullbackLeibler_numba_kl_div_has_inf(self):
        x = self.u1 * self.mask
        x *= -1
        
        from cil.optimisation.functions.KullbackLeibler import kl_div

        numpy.testing.assert_equal(kl_div(x.array, self.b1.array, self.eta.array), numpy.inf)

    def test_KullbackLeibler_numba_kl_div_mask_has_inf(self):
        x = self.u1 * self.mask
        x *= -1
        
        from cil.optimisation.functions.KullbackLeibler import kl_div_mask

        numpy.testing.assert_equal(kl_div_mask(x.array, self.b1.array, self.eta.array, self.mask.array), numpy.inf)

    def test_KullbackLeibler_numba_call_mask(self):
        f = self.f
        f_np = self.f_np
        tau = self.tau
        u1 = self.u1
        g1 = self.g1
        mask = self.mask

        u1_on_mask = self.u1_on_mask
        f_on_mask = self.f_on_mask
        f_mask = self.f_mask
        f_mask_c = self.f_mask_c

        numpy.testing.assert_allclose(f_mask(u1) + f_mask_c(u1), f(u1),  rtol=1e-5)

    def test_KullbackLeibler_numba_proximal(self):
        f = self.f
        f_np = self.f_np
        tau = self.tau
        u1 = self.u1

        numpy.testing.assert_allclose(f.proximal(u1,tau=tau).as_array(),
                                      f_np.proximal(u1,tau=tau).as_array(), rtol=7e-3)
        numpy.testing.assert_array_almost_equal(f.proximal(u1,tau=tau).as_array(),
        f_np.proximal(u1,tau=tau).as_array(), decimal=4)

    def test_KullbackLeibler_numba_proximal_arr(self):
        f = self.f
        f_np = self.f_np
        tau = self.u1.copy()
        tau.fill(self.tau)
        u1 = self.u1
        a = f.proximal(u1,tau=self.tau)
        b = f.proximal(u1,tau=tau)
        numpy.testing.assert_allclose(f.proximal(u1,tau=self.tau).as_array(),
                                      f.proximal(u1,tau=tau).as_array(), rtol=7e-3)
        numpy.testing.assert_array_almost_equal(f.proximal(u1,tau=self.tau).as_array(),
                                                f.proximal(u1,tau=tau).as_array(), decimal=4)

    def test_KullbackLeibler_numba_gradient(self):
        with self.subTest():
            f = self.f
            f_np = self.f_np
            u1 = self.u1
            grad_np = f_np.gradient(u1).as_array()
            numpy.testing.assert_allclose(f.gradient(u1).as_array(), grad_np, rtol=1e-3)
        with self.subTest("mask"):
            f = self.f_mask
            mask = self.mask>0
            grad_mask = f.gradient(u1).as_array()
            numpy.testing.assert_allclose(grad_mask[mask], grad_np[mask], rtol=1e-3)
            self.assertFalse(grad_mask[~mask].any())

    def test_KullbackLeibler_numba_convex_conjugate(self):
        f = self.f
        f_np = self.f_np
        tau = self.tau
        u1 = self.u1

        numpy.testing.assert_allclose(f.convex_conjugate(u1), f_np.convex_conjugate(u1), rtol=1e-3)

    def test_KullbackLeibler_numba_proximal_conjugate_arr(self):
        f = self.f
        f_np = self.f_np
        tau = self.tau
        u1 = self.u1

        numpy.testing.assert_allclose(f.proximal_conjugate(u1,tau=tau).as_array(),
                        f_np.proximal_conjugate(u1,tau=tau).as_array(), rtol=1e-3)

    def test_KullbackLeibler_numba_convex_conjugate_mask(self):
        f = self.f
        tau = self.tau
        u1 = self.u1

        mask = self.mask
        f_mask = self.f_mask
        f_mask_c = self.f_mask_c
        f_on_mask = self.f_on_mask
        u1_on_mask = self.u1_on_mask

        numpy.testing.assert_allclose(
            f.convex_conjugate(u1),
            f_mask.convex_conjugate(u1) + f_mask_c.convex_conjugate(u1) ,\
                 rtol=1e-3)

    def test_KullbackLeibler_numba_proximal_conjugate_mask(self):
        f = self.f
        f_mask = self.f_mask
        f_mask_c = self.f_mask_c
        x = self.u1
        m = self.mask
        m_c = self.mask_c
        tau = self.tau

        out = x * 0
        out_c = x * 0
        f_mask_c.proximal_conjugate(x,tau=tau, out=out_c)
        f_mask.proximal_conjugate(x,tau=tau, out=out)
        numpy.testing.assert_allclose(f.proximal_conjugate(x,tau=tau).as_array(),
                                      (out + out_c).as_array(), rtol=7e-3)
        # print ("f.prox_conj\n"       , f.proximal_conjugate(x,tau=tau).as_array())
        # print ("f_mask.prox_conj\n"  , out.as_array())
        # print ("f_mask_c.prox_conj\n", out_c.as_array())
        b = f_mask_c.proximal_conjugate(x,tau=tau)
        a = f_mask.proximal_conjugate(x,tau=tau)
        numpy.testing.assert_allclose(f.proximal_conjugate(x,tau=tau).as_array(),
                                      (f_mask.proximal_conjugate(x,tau=tau) +\
                                      f_mask_c.proximal_conjugate(x, tau=tau)) .as_array(), rtol=7e-3)



#  Copyright 2018 United Kingdom Research and Innovation
#  Copyright 2018 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
import numpy

from .labels import AcquisitionDimension, Backend
from .data_container import DataContainer
from .partitioner import Partitioner


class AcquisitionData(DataContainer, Partitioner):
    '''DataContainer for holding 2D or 3D sinogram'''
    __container_priority__ = 1

    @property
    def geometry(self):
        return self._geometry

    @geometry.setter
    def geometry(self, val):
        self._geometry = val

    @property
    def dimension_labels(self):
        return self.geometry.dimension_labels

    @dimension_labels.setter
    def dimension_labels(self, val):
        if val is not None:
            raise ValueError("Unable to set the dimension_labels directly. Use geometry.set_labels() instead")

    def __init__(self,
                 array = None,
                 deep_copy=True,
                 geometry = None,
                 **kwargs):

        dtype = kwargs.get('dtype', numpy.float32)

        if geometry is None:
            raise AttributeError("AcquisitionData requires a geometry")

        labels = kwargs.get('dimension_labels', None)
        if labels is not None and labels != geometry.dimension_labels:
                raise ValueError("Deprecated: 'dimension_labels' cannot be set with 'allocate()'. Use 'geometry.set_labels()' to modify the geometry before using allocate.")

        if array is None:
            array = numpy.empty(geometry.shape, dtype=dtype)
        elif issubclass(type(array) , DataContainer):
            array = array.as_array()
        elif issubclass(type(array) , numpy.ndarray):
            # remove singleton dimensions
            array = numpy.squeeze(array)
        else:
            raise TypeError('array must be a CIL type DataContainer or numpy.ndarray got {}'.format(type(array)))

        if array.shape != geometry.shape:
            raise ValueError('Shape mismatch got {} expected {}'.format(array.shape, geometry.shape))

        super(AcquisitionData, self).__init__(array, deep_copy, geometry=geometry,**kwargs)

    def __eq__(self, other):
        '''
        Check if two AcquisitionData objects are equal. This is done by checking if the geometry, data and dtype are equal.
        Also, if the other object is a numpy.ndarray, it will check if the data and dtype are equal.
        
        Parameters
        ----------
        other: AcquisitionData or numpy.ndarray
            The object to compare with.
        
        Returns
        -------
        bool
            True if the two objects are equal, False otherwise.
        '''

        if isinstance(other, AcquisitionData):
            if numpy.array_equal(self.as_array(), other.as_array()) \
                and self.geometry == other.geometry \
                and self.dtype == other.dtype:
                return True 
        elif numpy.array_equal(self.as_array(), other) and self.dtype==other.dtype:
            return True
        else:
            return False

    def get_slice(self,channel=None, angle=None, vertical=None, horizontal=None, force=False):
        '''Returns a new dataset of a single slice in the requested direction.'''
        try:
            geometry_new = self.geometry.get_slice(channel=channel, angle=angle, vertical=vertical, horizontal=horizontal)
        except ValueError:
            if force:
                geometry_new = None
            else:
                raise ValueError ("Unable to return slice of requested AcquisitionData. Use 'force=True' to return DataContainer instead.")

        #get new data
        #if vertical = 'centre' slice convert to index and subset, this will interpolate 2 rows to get the center slice value
        if vertical == 'centre':
            dim = self.geometry.dimension_labels.index('vertical')

            centre_slice_pos = (self.geometry.shape[dim]-1) / 2.
            ind0 = int(numpy.floor(centre_slice_pos))
            w2 = centre_slice_pos - ind0
            out = DataContainer.get_slice(self, channel=channel, angle=angle, vertical=ind0, horizontal=horizontal)

            if w2 > 0:
                out2 = DataContainer.get_slice(self, channel=channel, angle=angle, vertical=ind0 + 1, horizontal=horizontal)
                out = out * (1 - w2) + out2 * w2
        else:
            out = DataContainer.get_slice(self, channel=channel, angle=angle, vertical=vertical, horizontal=horizontal)

        if len(out.shape) == 1 or geometry_new is None:
            return out
        else:
            return AcquisitionData(out.array, deep_copy=False, geometry=geometry_new, suppress_warning=True)

    def reorder(self, order):
        '''
        Reorders the data in memory as requested. This is an in-place operation.

        Parameters
        ----------
        order: list or str
            Ordered list of labels from self.dimension_labels, or string 'astra' or 'tigre'.
        '''
        if order in Backend:
            order = AcquisitionDimension.get_order_for_engine(order, self.geometry)

        super().reorder(order)



#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import os
import shutil
import unittest

import numpy as np

from cil.framework import BlockDataContainer
from cil.optimisation.algorithms import FISTA, ISTA
from cil.optimisation.functions import TotalVariation, L2NormSquared, KullbackLeibler, IndicatorBox
from cil.optimisation.operators import GradientOperator, LinearOperator
from cil.optimisation.utilities import AdaptiveSensitivity
from testclass import CCPiTestClass
from utils import has_ccpi_regularisation, initialise_tests

initialise_tests()

try:
    import sirf.STIR as pet
    import sirf.Gadgetron as mr
    import sirf.Reg as reg
    from sirf.Utilities import examples_data_path

    has_sirf = True
except ImportError as ie:
    has_sirf = False

if has_ccpi_regularisation:
    from cil.plugins.ccpi_regularisation.functions import FGP_TV, TGV, FGP_dTV, TNV



class KullbackLeiblerSIRF(object):

    def setUp(self):

        if has_sirf:
            self.image1 = pet.ImageData(os.path.join(
                examples_data_path('PET'),'thorax_single_slice','emission.hv')
                )

            self.eta = self.image1.get_uniform_copy(0.1)
            self.x = self.image1.get_uniform_copy(0.4)

        self.f_np = KullbackLeibler(b = self.image1, backend='numpy')
        self.f1_np = KullbackLeibler(b = self.image1, eta = self.eta,  backend='numpy')
        self.out_np = self.image1.get_uniform_copy(0.)
        self.out_nb = self.image1.get_uniform_copy(0.)

        self.f_nb = KullbackLeibler(b = self.image1, backend='numba')
        self.f1_nb = KullbackLeibler(b = self.image1, eta = self.eta,  backend='numba')
        self.out1_np = self.image1.get_uniform_copy(0.)
        self.out1_nb = self.image1.get_uniform_copy(0.)

        self.tau = 400.4

    def tearDown(self):
        pass

    @unittest.skipUnless(has_sirf, "Skipping as SIRF is not available")
    def test_KullbackLeibler_call(self):
        np.testing.assert_almost_equal(self.f_np(self.x), self.f_nb(self.x), decimal = 2)
        np.testing.assert_almost_equal(self.f1_np(self.x), self.f1_nb(self.x), decimal = 2)


    @unittest.skipUnless(has_sirf, "Skipping as SIRF is not available")
    def test_KullbackLeibler_gradient(self):

        self.f_np.gradient(self.x, out = self.out_np)
        self.f_nb.gradient(self.x, out = self.out_nb)
        self.f1_np.gradient(self.x, out = self.out1_np)
        self.f1_nb.gradient(self.x, out = self.out1_nb)

        np.testing.assert_array_almost_equal(self.out_np.as_array(), self.out_nb.as_array(), decimal = 2)
        np.testing.assert_array_almost_equal(self.out1_np.as_array(), self.out1_nb.as_array(), decimal = 2)


    @unittest.skipUnless(has_sirf, "Skipping as SIRF is not available")
    def test_KullbackLeibler_convex_conjugate(self):

        np.testing.assert_almost_equal(self.f_np.convex_conjugate(self.x), self.f_nb.convex_conjugate(self.x), decimal = 2)
        np.testing.assert_almost_equal(self.f1_np.convex_conjugate(self.x), self.f1_nb.convex_conjugate(self.x), decimal = 2)


    @unittest.skipUnless(has_sirf, "Skipping as SIRF is not available")
    def test_KullbackLeibler_proximal(self):

        self.f_np.proximal(self.x, self.tau, out = self.out_np)
        self.f_nb.proximal(self.x, self.tau, out = self.out_nb)
        self.f1_np.proximal(self.x, self.tau, out = self.out1_np)
        self.f1_nb.proximal(self.x, self.tau, out = self.out1_nb)

        np.testing.assert_array_almost_equal(self.out_np.as_array(), self.out_nb.as_array(), decimal = 2)
        np.testing.assert_array_almost_equal(self.out1_np.as_array(), self.out1_nb.as_array(), decimal = 2)


    @unittest.skipUnless(has_sirf, "Skipping as SIRF is not available")
    def test_KullbackLeibler_proximal_conjugate(self):

        self.f_np.proximal_conjugate(self.x, self.tau, out = self.out_np)
        self.f_nb.proximal_conjugate(self.x, self.tau, out = self.out_nb)
        self.f1_np.proximal_conjugate(self.x, self.tau, out = self.out1_np)
        self.f1_nb.proximal_conjugate(self.x, self.tau, out = self.out1_nb)

        np.testing.assert_array_almost_equal(self.out_np.as_array(), self.out_nb.as_array(), decimal = 2)
        np.testing.assert_array_almost_equal(self.out1_np.as_array(), self.out1_nb.as_array(), decimal = 2)


class GradientSIRF(object):

    @unittest.skipUnless(has_sirf, "Skipping as SIRF is not available")
    def test_Gradient(self):

        #######################################
        ##### Test Gradient numpy backend #####
        #######################################

        Grad_numpy = GradientOperator(self.image1, backend='numpy')

        res1 = Grad_numpy.direct(self.image1)
        res2 = Grad_numpy.range_geometry().allocate()
        Grad_numpy.direct(self.image1, out=res2)

        self.assertTrue(isinstance(res1,BlockDataContainer))
        self.assertTrue(isinstance(res2,BlockDataContainer))

        for i in range(len(res1)):

            if isinstance(self.image1, pet.ImageData):
                self.assertTrue(isinstance(res1[i], pet.ImageData))
                self.assertTrue(isinstance(res2[i], pet.ImageData))
            else:
                self.assertTrue(isinstance(res1[i], mr.ImageData))
                self.assertTrue(isinstance(res2[i], mr.ImageData))
            # test direct with and without out
            np.testing.assert_array_almost_equal(res1[i].as_array(), res2[i].as_array())

        # test adjoint with and without out
        res3 = Grad_numpy.adjoint(res1)
        res4 = Grad_numpy.domain_geometry().allocate()
        Grad_numpy.adjoint(res2, out=res4)
        np.testing.assert_array_almost_equal(res3.as_array(), res4.as_array())

        # test dot_test
        for sd in [5,10]:

            self.assertTrue(LinearOperator.dot_test(Grad_numpy, seed=sd))

        # test shape of output of direct

        # check in the case of pseudo 2D data, e.g., (1, 155, 155)
        if 1 in self.image1.shape:
            self.assertEqual(res1.shape, (2,1))
        else:
            self.assertEqual(res1.shape, (3,1))

        ########################################
        ##### Test Gradient c backend  #####
        ########################################
        Grad_c = GradientOperator(self.image1, backend='c')

        # test direct with and without out
        res5 = Grad_c.direct(self.image1)
        res6 = Grad_c.range_geometry().allocate()*0.
        Grad_c.direct(self.image1, out=res6)

        for i in range(len(res5)):
            np.testing.assert_array_almost_equal(res5[i].as_array(), res6[i].as_array())

            # compare c vs numpy gradient backends
            np.testing.assert_array_almost_equal(res6[i].as_array(), res2[i].as_array())


        # test dot_test
        for sd in [5,10]:
            self.assertTrue(LinearOperator.dot_test(Grad_c, seed=sd))

        # test adjoint
        res7 = Grad_c.adjoint(res5)
        res8 = Grad_c.domain_geometry().allocate()*0.
        Grad_c.adjoint(res5, out=res8)
        np.testing.assert_array_almost_equal(res7.as_array(), res8.as_array())



class TestGradientPET_2D(unittest.TestCase, GradientSIRF):

    def setUp(self):

        if has_sirf:
            self.image1 = pet.ImageData(os.path.join(
                examples_data_path('PET'),'thorax_single_slice','emission.hv')
                )

    def tearDown(self):
        pass

class TestGradientPET_3D(unittest.TestCase, GradientSIRF):

    def setUp(self):

        if has_sirf:
            self.image1 = pet.ImageData(os.path.join(
                examples_data_path('PET'),'brain','emission.hv')
                )

    def tearDown(self):
        pass

class TestGradientMR_2D(unittest.TestCase, GradientSIRF):

    def setUp(self):

        if has_sirf:
            acq_data = mr.AcquisitionData(os.path.join
                (examples_data_path('MR'),'simulated_MR_2D_cartesian.h5')
            )
            preprocessed_data = mr.preprocess_acquisition_data(acq_data)
            recon = mr.FullySampledReconstructor()
            recon.set_input(preprocessed_data)
            recon.process()
            self.image1 = recon.get_output()

    def tearDown(self):
        pass

    @unittest.skipUnless(has_sirf, "Has SIRF")
    def test_TVdenoisingMR(self):

        # compare inplace proximal method of TV
        alpha = 0.5
        TV = alpha * TotalVariation(max_iteration=10, warm_start=False)
        res1 = TV.proximal(self.image1, tau=1.0)

        res2 = self.image1*0.
        TV.proximal(self.image1, tau=1.0, out=res2)
        np.testing.assert_array_almost_equal(res1.as_array(), res2.as_array(), decimal=3)

        # compare with FISTA algorithm
        f =  0.5 * L2NormSquared(b=self.image1)
        fista = FISTA(initial=self.image1*0.0, f=f, g=TV, max_iteration=10, update_objective_interval=10)
        fista.run(verbose=0)
        np.testing.assert_array_almost_equal(fista.solution.as_array(), res2.as_array(), decimal=3)


class TestSIRFCILIntegration(CCPiTestClass):

    def setUp(self):
        if has_sirf:
            os.chdir(examples_data_path('PET'))
            # Copy files to a working folder and change directory to where these files are.
            # We do this to avoid cluttering your SIRF files. This way, you can delete
            # working_folder and start from scratch.
            shutil.rmtree('working_folder/brain',True)
            shutil.copytree('brain','working_folder/brain')
            os.chdir('working_folder/brain')

            self.cwd = os.getcwd()


    def tearDown(self):
        if has_sirf:
            shutil.rmtree(self.cwd)


    @unittest.skipUnless(has_sirf, "Has SIRF")
    def test_BlockDataContainer_with_SIRF_DataContainer_divide(self):
        os.chdir(self.cwd)
        image1 = pet.ImageData('emission.hv')
        image2 = pet.ImageData('emission.hv')
        image1.fill(1.)
        image2.fill(2.)

        tmp = image1.divide(1.)
        np.testing.assert_array_equal(image1.as_array(), tmp.as_array())
        tmp = image2.divide(1.)
        np.testing.assert_array_equal(image2.as_array(), tmp.as_array())


        # image.fill(1.)
        bdc = BlockDataContainer(image1, image2)
        bdc1 = bdc.divide(1.)

        # self.assertBlockDataContainerEqual(bdc , bdc1)
        np.testing.assert_allclose(bdc.get_item(0).as_array(), bdc1.get_item(0).as_array())
        np.testing.assert_allclose(bdc.get_item(1).as_array(), bdc1.get_item(1).as_array())


    @unittest.skipUnless(has_sirf, "Has SIRF")
    def test_BlockDataContainer_with_SIRF_DataContainer_multiply(self):
        os.chdir(self.cwd)
        image1 = pet.ImageData('emission.hv')
        image2 = pet.ImageData('emission.hv')
        image1.fill(1.)
        image2.fill(2.)

        tmp = image1.multiply(1.)
        np.testing.assert_array_equal(image1.as_array(), tmp.as_array())
        tmp = image2.multiply(1.)
        np.testing.assert_array_equal(image2.as_array(), tmp.as_array())


        # image.fill(1.)
        bdc = BlockDataContainer(image1, image2)
        bdc1 = bdc.multiply(1.)

        # self.assertBlockDataContainerEqual(bdc , bdc1)
        np.testing.assert_allclose(bdc.get_item(0).as_array(), bdc1.get_item(0).as_array())
        np.testing.assert_allclose(bdc.get_item(1).as_array(), bdc1.get_item(1).as_array())


    @unittest.skipUnless(has_sirf, "Has SIRF")
    def test_BlockDataContainer_with_SIRF_DataContainer_add(self):
        os.chdir(self.cwd)
        image1 = pet.ImageData('emission.hv')
        image2 = pet.ImageData('emission.hv')
        image1.fill(0)
        image2.fill(1)

        tmp = image1.add(1.)
        np.testing.assert_array_equal(image2.as_array(), tmp.as_array())
        tmp = image2.subtract(1.)
        np.testing.assert_array_equal(image1.as_array(), tmp.as_array())

        bdc = BlockDataContainer(image1, image2)
        bdc1 = bdc.add(1.)

        image1.fill(1)
        image2.fill(2)

        bdc = BlockDataContainer(image1, image2)

        np.testing.assert_allclose(bdc.get_item(0).as_array(), bdc1.get_item(0).as_array())
        np.testing.assert_allclose(bdc.get_item(1).as_array(), bdc1.get_item(1).as_array())
        # self.assertBlockDataContainerEqual(bdc , bdc1)


    @unittest.skipUnless(has_sirf, "Has SIRF")
    def test_BlockDataContainer_with_SIRF_DataContainer_subtract(self):
        os.chdir(self.cwd)
        image1 = pet.ImageData('emission.hv')
        image2 = pet.ImageData('emission.hv')
        image1.fill(2)
        image2.fill(1)


        bdc = BlockDataContainer(image1, image2)
        bdc1 = bdc.subtract(1.)

        image1.fill(1)
        image2.fill(0)

        bdc = BlockDataContainer(image1, image2)

        # self.assertBlockDataContainerEqual(bdc , bdc1)
        np.testing.assert_allclose(bdc.get_item(0).as_array(), bdc1.get_item(0).as_array())
        np.testing.assert_allclose(bdc.get_item(1).as_array(), bdc1.get_item(1).as_array())



class CCPiRegularisationWithSIRFTests():

    def setUpFGP_TV(self, max_iteration=100, alpha=1.):
        return alpha*FGP_TV(max_iteration=max_iteration)

    @unittest.skipUnless(has_sirf and has_ccpi_regularisation, "Has SIRF and CCPi Regularisation")
    def test_FGP_TV_call_works(self):
        regulariser = self.setUpFGP_TV()
        output_number = regulariser(self.image1)
        self.assertTrue(True)
        # TODO: test the actual value
        # expected = 160600016.0
        # np.testing.assert_allclose(output_number, expected, rtol=1e-5)

    @unittest.skipUnless(has_sirf and has_ccpi_regularisation, "Has SIRF and CCPi Regularisation")
    def test_FGP_TV_proximal_works(self):
        regulariser = self.setUpFGP_TV()
        solution = regulariser.proximal(x=self.image1, tau=1)
        self.assertTrue(True)

    # TGV
    def setUpTGV(self, max_iteration=100, alpha=1.):
        return alpha * TGV(max_iteration=max_iteration)

    @unittest.skipUnless(has_sirf and has_ccpi_regularisation, "Has SIRF and CCPi Regularisation")
    def test_TGV_call_works(self):
        regulariser = self.setUpTGV()
        output_number = regulariser(self.image1)
        self.assertTrue(True)

    @unittest.skipUnless(has_sirf and has_ccpi_regularisation, "Has SIRF and CCPi Regularisation")
    def test_TGV_proximal_works(self):
        regulariser = self.setUpTGV()
        solution = regulariser.proximal(x=self.image1, tau=1)
        self.assertTrue(True)

    # dTV
    def setUpdTV(self, max_iteration=100, alpha=1.):
        return alpha * FGP_dTV(reference=self.image2, max_iteration=max_iteration)

    @unittest.skipUnless(has_sirf and has_ccpi_regularisation, "Has SIRF and CCPi Regularisation")
    def test_TGV_call_works(self):
        regulariser = self.setUpTGV()
        output_number = regulariser(self.image1)
        self.assertTrue(True)

    @unittest.skipUnless(has_sirf and has_ccpi_regularisation, "Has SIRF and CCPi Regularisation")
    def test_TGV_proximal_works(self):
        regulariser = self.setUpTGV()
        solution = regulariser.proximal(x=self.image1, tau=1)
        self.assertTrue(True)

    # TNV
    def setUpTNV(self, max_iteration=100, alpha=1.):
        return alpha * TNV(max_iteration=max_iteration)

    @unittest.skipUnless(has_sirf and has_ccpi_regularisation, "Has SIRF and CCPi Regularisation")
    def test_TNV_call_works(self):
        new_shape = [ i for i in self.image1.shape if i!=1]
        if len(new_shape) == 3:
            regulariser = self.setUpTNV()
            output_number = regulariser(self.image1)
            self.assertTrue(True)

    @unittest.skipUnless(has_sirf and has_ccpi_regularisation, "Has SIRF and CCPi Regularisation")
    def test_TNV_proximal_works(self):
        new_shape = [ i for i in self.image1.shape if i!=1]
        if len(new_shape) == 3:
            regulariser = self.setUpTNV()
            solution = regulariser.proximal(x=self.image1, tau=1.)
            self.assertTrue(True)

class TestPETRegularisation(unittest.TestCase, CCPiRegularisationWithSIRFTests):
    skip_TNV_on_2D = True
    def setUp(self):
        self.image1 = pet.ImageData(os.path.join(
            examples_data_path('PET'),'thorax_single_slice','emission.hv'
            ))
        self.image2 = self.image1 * 0.5

    @unittest.skipIf(skip_TNV_on_2D, "TNV not implemented for 2D")
    def test_TNV_call_works(self):
        super().test_TNV_call_works()

    @unittest.skipIf(skip_TNV_on_2D, "TNV not implemented for 2D")
    def test_TNV_proximal_works(self):
        super().test_TNV_proximal_works()

class TestRegRegularisation(unittest.TestCase, CCPiRegularisationWithSIRFTests):
    def setUp(self):
        self.image1 = reg.ImageData(os.path.join(examples_data_path('Registration'),'test2.nii.gz'))
        self.image2 = self.image1 * 0.5

class TestMRRegularisation(unittest.TestCase, CCPiRegularisationWithSIRFTests):
    def setUp(self):
        acq_data = mr.AcquisitionData(os.path.join(examples_data_path('MR'),'simulated_MR_2D_cartesian.h5'))
        preprocessed_data = mr.preprocess_acquisition_data(acq_data)
        recon = mr.FullySampledReconstructor()
        recon.set_input(preprocessed_data)
        recon.process()
        self.image1 = recon.get_output()
        self.image2 = self.image1 * 0.5


class TestCILSIRFPrecond(unittest.TestCase):

    @unittest.skipUnless(has_sirf, "Skipping as SIRF is not available")
    def setUp(self):
        data_path = os.path.join(examples_data_path('PET'), 'thorax_single_slice')
        self.image = pet.ImageData(os.path.join(data_path, 'emission.hv'))
        attn_image = pet.ImageData(os.path.join(data_path, 'attenuation.hv'))
        template = pet.AcquisitionData(os.path.join(data_path, 'template_sinogram.hs'))
        self.cmax = self.image.max()*.6
        # create attenuation
        acq_model_for_attn = pet.AcquisitionModelUsingRayTracingMatrix()
        asm_attn = pet.AcquisitionSensitivityModel(attn_image, acq_model_for_attn)
        asm_attn.set_up(template)
        attn_factors = asm_attn.forward(template.get_uniform_copy(1))
        asm_attn = pet.AcquisitionSensitivityModel(attn_factors)

        # create acquisition model
        self.acq_model = pet.AcquisitionModelUsingRayTracingMatrix()
        self.acq_model.set_num_tangential_LORs(5)
        self.acq_model.set_acquisition_sensitivity(asm_attn)
        self.acq_model.set_up(template,self.image)    

        # simulate data
        acquired_data = self.acq_model.forward(self.image)

        background_term = acquired_data.get_uniform_copy(acquired_data.max()/10)
        self.acq_model.set_background_term(background_term)
        self.acquired_data = self.acq_model.forward(self.image)

    @unittest.skipUnless(has_sirf, "Skipping as SIRF is not available")
    def test_SIRF_CIL_MLEM(self):     

        obj_fun = pet.make_Poisson_loglikelihood(self.acquired_data)
        obj_fun.set_acquisition_model(self.acq_model)
        recon = pet.OSMAPOSLReconstructor()
        recon.set_objective_function(obj_fun)
        recon.set_num_subsets(1)
        recon.set_num_subiterations(10)

        initial_image=self.image.get_uniform_copy(self.cmax / 4)
        trunc = pet.TruncateToCylinderProcessor()
        trunc.apply(initial_image)

        recon.set_current_estimate(initial_image)
        recon.set_up(initial_image)
        recon.process()
        reconstructed_image=recon.get_output()    

        sens = AdaptiveSensitivity(self.acq_model, delta = 0., max_iterations = 200)

        trunc = pet.TruncateToCylinderProcessor()
        trunc.apply(initial_image)

        ista = ISTA(initial = initial_image, f = obj_fun, g = IndicatorBox(lower=0.),  step_size = -1.0, 
                    preconditioner=sens,update_objective_interval=1, max_iteration=10)
        ista.run(verbose=0)      

        print("adad")
        np.testing.assert_allclose(ista.solution.as_array()[0], reconstructed_image.as_array()[0], rtol=1e-3)         



#  Copyright 2024 United Kingdom Research and Innovation
#  Copyright 2024 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import unittest


import numpy as np

from cil.utilities.errors import InPlaceError
from cil.utilities import dataexample
from cil.framework import AcquisitionGeometry, ImageGeometry, VectorGeometry, DataContainer
from cil.framework.labels import AcquisitionType

from cil.optimisation.operators import IdentityOperator, WaveletOperator
from cil.optimisation.functions import  KullbackLeibler, ConstantFunction, TranslateFunction, soft_shrinkage, L1Sparsity, BlockFunction
from cil.optimisation.operators import LinearOperator, MatrixOperator  

from cil.optimisation.operators import SumOperator,  ZeroOperator, CompositionOperator, ProjectionMap
from cil.optimisation.operators import BlockOperator,\
    FiniteDifferenceOperator, SymmetrisedGradientOperator,  DiagonalOperator, MaskOperator, ChannelwiseOperator, BlurringOperator

from cil.optimisation.functions import  KullbackLeibler, WeightedL2NormSquared, L2NormSquared, \
    L1Norm, L2NormSquared, MixedL21Norm, LeastSquares, \
    SmoothMixedL21Norm, OperatorCompositionFunction, \
     IndicatorBox, TotalVariation,  SumFunction, SumScalarFunction, \
    WeightedL2NormSquared, MixedL11Norm, ZeroFunction

from cil.processors import AbsorptionTransmissionConverter, Binner, CentreOfRotationCorrector, MaskGenerator, Masker, Normaliser, Padder, \
RingRemover, Slicer, TransmissionAbsorptionConverter, PaganinProcessor, FluxNormaliser

import numpy
from utils import has_tigre, has_nvidia


from cil.framework import  BlockGeometry
from cil.optimisation.functions import TranslateFunction
from timeit import default_timer as timer

import numpy as np


from testclass import CCPiTestClass
from cil.utilities.quality_measures import mae


from utils import  initialise_tests



initialise_tests()


class TestFunctionOutAndInPlace(CCPiTestClass):

    def setUp(self):

        ag = AcquisitionGeometry.create_Parallel2D()
        angles = np.linspace(0, 360, 10, dtype=np.float32)

        # default
        ag.set_angles(angles)
        ag.set_panel(10)

        ig = ag.get_ImageGeometry()

        scalar = 4

        b = ag.allocate('random', seed=2)
        weight_ls = ig.allocate('random', seed=2)

        A = IdentityOperator(ig)
        b_ig = ig.allocate('random')
        c = numpy.float64(0.3)
        bg = BlockGeometry(ig, ig)
        # [(function, geometry, test_proximal, test_proximal_conjugate, test_gradient), ...]
        self.func_geom_test_list = [
            (IndicatorBox(), ag, True, True, False),
            (KullbackLeibler(b=b, backend='numba'), ag, True, True, True),
            (KullbackLeibler(b=b, backend='numpy'), ag, True, True, True),
            (L1Norm(), ag, True, True, False),
            (L1Norm(), ig, True, True, False),
            (L1Norm(b=b), ag, True, True, False),
            (L1Norm(b=b, weight=b), ag, True, True, False),
            (TranslateFunction(L1Norm(), b), ag, True, True, False),
            (TranslateFunction(L2NormSquared(), b), ag, True, True, True),
            (L2NormSquared(), ag, True, True, True),
            (scalar * L2NormSquared(), ag, True, True, True),
            (SumFunction(L2NormSquared(), scalar * L2NormSquared()), ag, False, False, True),
            (SumScalarFunction(L2NormSquared(), 3), ag, True, True, True),
            (ConstantFunction(3), ag, True, True, True),
            (ZeroFunction(), ag, True, True, True),
            (L2NormSquared(b=b), ag, True, True, True),
            (L2NormSquared(), ag, True, True, True),
            (LeastSquares(A, b_ig, c, weight_ls), ig, False, False, True),
            (LeastSquares(A, b_ig, c), ig, False, False, True),
            (WeightedL2NormSquared(weight=b_ig), ig, True, True, True),
            (TotalVariation(backend='c', warm_start=False, max_iteration=100), ig, True, True, False),
            (TotalVariation(backend='numpy', warm_start=False, max_iteration=100), ig, True, True, False),
            (OperatorCompositionFunction(L2NormSquared(), A), ig, False, False, True),
            (MixedL21Norm(), bg, True, True, False),
            (SmoothMixedL21Norm(epsilon=0.3), bg, False, False, True),
            (MixedL11Norm(), bg, True, True, False),
            (BlockFunction(L1Norm(),L2NormSquared()), bg, True, True, False),
            (BlockFunction(L2NormSquared(),L2NormSquared()), bg, True, True, True),
            (L1Sparsity(WaveletOperator(ig)), ig, True, True, False)


        ]

        np.random.seed(5)
        self.data_arrays=[np.random.normal(0,1, (10,10)).astype(np.float32),  np.array(range(0,65500, 655), dtype=np.uint16).reshape((10,10)), np.random.uniform(-0.1,1,(10,10)).astype(np.float32)]

    def get_result(self, function, method, x, *args):
        try:
            input=x.copy() #To check that it isn't changed after function calls
            if method == 'proximal':
                out= function.proximal(x, *args)
            elif method == 'proximal_conjugate':
                out= function.proximal_conjugate(x, *args)
            elif method == 'gradient':
                out= function.gradient(x, *args)
            self.assertDataArraysInContainerAllClose(input, x, rtol=1e-5, msg= "In case func."+method+'(data, *args) where func is  ' + function.__class__.__name__+ 'the input data has been incorrectly affected by the calculation. ')
            return out
        except NotImplementedError:
            raise NotImplementedError(function.__class__.__name__+" raises a NotImplementedError for "+method)


    def in_place_test(self,desired_result, function, method,   x, *args, ):
            out3 = x.copy()
            try:
                try:
                    if method == 'proximal':
                        function.proximal(out3, *args, out=out3)
                    elif method == 'proximal_conjugate':
                        function.proximal_conjugate(out3, *args, out=out3)
                    elif method == 'gradient':
                        function.gradient(out3, *args, out=out3)
                    self.assertDataArraysInContainerAllClose(desired_result, out3, rtol=1e-5, msg= "In place calculation failed for func."+method+'(data, *args, out=data) where func is  ' + function.__class__.__name__+ '. ')

                except InPlaceError:
                    pass
            except NotImplementedError:
                raise NotImplementedError(function.__class__.__name__+" raises a NotImplementedError for "+method)

    def out_test(self, desired_result, function, method,  x, *args, ):
        input = x.copy()
        out2=0*(x.copy())
        try:
            try:
                if method == 'proximal':
                    ret = function.proximal(input, *args, out=out2)
                elif method == 'proximal_conjugate':
                    ret = function.proximal_conjugate(input, *args, out=out2)
                elif method == 'gradient':
                    ret = function.gradient(input, *args, out=out2)
                self.assertDataArraysInContainerAllClose(desired_result, out2, rtol=1e-5, msg= "Calculation failed using `out` in func."+method+'(x, *args, out=data) where func is  ' + function.__class__.__name__+ '. ')
                self.assertDataArraysInContainerAllClose(input, x,  rtol=1e-5, msg= "In case func."+method+'(data, *args, out=out) where func is  ' + function.__class__.__name__+ 'the input data has been incorrectly affected by the calculation. ')
                self.assertDataArraysInContainerAllClose(desired_result, ret, rtol=1e-5, msg= f"Calculation failed returning with `out` in ret = func.{method}(x, *args, out=data) where func is {function.__class__.__name__}")
            
            except InPlaceError:
                pass
        except NotImplementedError:
            raise NotImplementedError(function.__class__.__name__+" raises a NotImplementedError for "+method)



    def test_proximal_conjugate_out(self):
        for func, geom, _, test_proximal_conj, _ in self.func_geom_test_list:
            if test_proximal_conj:
                for data_array in self.data_arrays:
                    data=geom.allocate(None)
                    data.fill(data_array)
                    result=self.get_result(func, 'proximal_conjugate', data, 0.5)
                    self.out_test(result, func,  'proximal_conjugate',  data, 0.5)
                    self.in_place_test(result, func, 'proximal_conjugate',  data, 0.5)

    def test_proximal_out(self):
        for func, geom, test_proximal, _, _  in self.func_geom_test_list:
            if test_proximal:
                for data_array in self.data_arrays:
                    data=geom.allocate(None)
                    data.fill(data_array)
                    result=self.get_result(func, 'proximal', data, 0.5)
                    self.out_test(result, func, 'proximal',  data, 0.5)
                    self.in_place_test(result,func,  'proximal',  data, 0.5)

    def test_gradient_out(self):
        for func, geom, _, _, test_gradient in self.func_geom_test_list:
            if test_gradient:
                for data_array in self.data_arrays:
                    print(func.__class__.__name__)
                    data=geom.allocate(None)
                    data.fill(data_array)
                    result=self.get_result(func, 'gradient', data)
                    self.out_test(result, func, 'gradient',   data)
                    self.in_place_test(result, func, 'gradient',   data)



class TestOperatorOutAndInPlace(CCPiTestClass):
    def setUp(self):

        ig = ImageGeometry(10,10,channels=3)
        ig_2D=ImageGeometry(10,10)
        vg = VectorGeometry(10)

        mask = ig.allocate(True,dtype=bool)
        amask = mask.as_array()
        amask[2,1:3,:] = False
        amask[0,0,:] = False




        # Parameters for point spread function PSF (size and std)
        ks          = 10
        ksigma      = 5.0

        # Create 1D PSF and 2D as outer product, then normalise.
        w           = numpy.exp(-numpy.arange(-(ks-1)/2,(ks-1)/2+1)**2/(2*ksigma**2))
        w.shape     = (ks,1)
        PSF         = w*numpy.transpose(w)
        PSF         = PSF/(PSF**2).sum()
        PSF         = PSF/PSF.sum()
        PSF         = np.array([PSF]*3)

        np.random.seed(5)

        self.operator_geom_test_list = [
            (MatrixOperator(numpy.random.randn(10, 10)), vg),
            (ZeroOperator(ig), ig),
            (IdentityOperator(ig), ig),
            (3 * IdentityOperator(ig), ig),
            (DiagonalOperator(ig.allocate('random',seed=101)), ig),
            (MaskOperator(mask), ig),
            (ChannelwiseOperator(DiagonalOperator(ig_2D.allocate('random',seed=101)),3), ig),
            (BlurringOperator(PSF,ig), ig),
            (FiniteDifferenceOperator(ig, direction = 0, bnd_cond = 'Neumann') , ig),
            (FiniteDifferenceOperator(ig, direction = 0) , ig)]
            


        self.data_arrays=[np.random.normal(0,1, (3,10,10)).astype(np.float32),  np.array(range(0,65400, 218), dtype=np.uint16).reshape((3,10,10)), np.random.uniform(-0.1,1,(3,10,10)).astype(np.float32)]
        self.vector_data_arrays=[np.random.normal(0,1, (10)).astype(np.float32),  np.array(range(0,65400, 6540), dtype=np.uint16), np.random.uniform(-0.1,1,(10)).astype(np.float32)]






    def get_result(self, operator, method, x, *args):
        try:
            input=x.copy() #To check that it isn't changed after function calls
            if method == 'direct':
                out= operator.direct(x, *args)
            elif method == 'adjoint':
                out= operator.adjoint(x, *args)

            self.assertDataArraysInContainerAllClose(input, x, rtol=1e-5, msg= "In case operator."+method+'(data, *args) where operator is  ' + operator.__class__.__name__+ 'the input data has been incorrectly affected by the calculation. ')
            return out
        except NotImplementedError:
            raise NotImplementedError(operator.__class__.__name__+" raises a NotImplementedError for "+method)

    def in_place_test(self,desired_result, operator, method,   x, *args, ):
            out3 = x.copy()
            try:
                try:
                    if method == 'direct':
                        operator.direct(out3, *args, out=out3)
                    elif method == 'adjoint':
                        operator.adjoint(out3, *args, out=out3)

                    self.assertDataArraysInContainerAllClose(desired_result, out3, rtol=1e-5, msg= "In place calculation failed for operator."+method+'(data, *args, out=data) where operator is  ' + operator.__class__.__name__+ '. ')

                except InPlaceError:
                    pass
            except NotImplementedError: 
                raise NotImplementedError(operator.__class__.__name__+" raises a NotImplementedError for "+method)


    def out_test(self, desired_result, operator, method,  x, *args):
        input = x.copy()
        out2=0*(x.copy())
        try:
            if method == 'direct':
                ret = operator.direct(input, *args, out=out2)
            elif method == 'adjoint':
                ret = operator.adjoint(input, *args, out=out2)

            self.assertDataArraysInContainerAllClose(desired_result, out2, rtol=1e-5, msg= "Calculation failed using `out` in operator."+method+'(x, *args, out=data) where func is  ' + operator.__class__.__name__+ '. ')
            self.assertDataArraysInContainerAllClose(input, x,  rtol=1e-5, msg= "In case operator."+method+'(data, *args, out=out) where operator is  ' + operator.__class__.__name__+ 'the input data has been incorrectly affected by the calculation. ')
            self.assertDataArraysInContainerAllClose(desired_result, ret, rtol=1e-5, msg= f"Calculation failed using return and `out` in ret = operator.{method}(x, *args, out=data) where func is {operator.__class__.__name__}")
            
        except (InPlaceError, NotImplementedError):
            pass

    def test_direct_out(self):
        for operator, geom in self.operator_geom_test_list:
            for data_array in self.data_arrays:
                data=geom.allocate(None)
                try:
                    data.fill(data_array)
                except:
                    data.fill(data_array[0,0,:])
                result=self.get_result(operator, 'direct', data)
                self.out_test(result, operator,  'direct',  data)
                self.in_place_test(result, operator, 'direct',  data)

    def test_proximal_out(self):
        for operator, geom in self.operator_geom_test_list:
            for data_array in self.data_arrays:
                data=geom.allocate(None)
                try:
                    data.fill(data_array)
                except:
                    data.fill(data_array[0,0,:])
                result=self.get_result(operator, 'adjoint', data)
                self.out_test(result, operator, 'adjoint',  data)
                self.in_place_test(result,operator,  'adjoint',  data)

class TestProcessorOutandInPlace(CCPiTestClass):
    def setUp(self):
        
        self.data_arrays=[np.random.normal(0,1, (10,20)).astype(np.float32),  
                          np.array(range(0,65500, 328), dtype=np.uint16).reshape((10,20)),
                          np.random.uniform(0,1,(10,20)).astype(np.float64)]

        ag_parallel_2D = AcquisitionGeometry.create_Parallel2D(detector_position=[0,1], units='m')
        angles = np.linspace(0, 360, 10, dtype=np.float32)
        ag_parallel_2D.set_angles(angles)
        ag_parallel_2D.set_panel(20)

        ag_parallel_3D = AcquisitionGeometry.create_Parallel3D(detector_position=[0,1,0], units='mm')
        # ag = AcquisitionGeometry.create_Parallel3D(detector_position=[0, 1, 0], units='mm').set_panel([20,2], pixel_size=0.1).set_angles(angles)
        ag_parallel_3D.set_angles(angles)
        ag_parallel_3D.set_panel([20,2], pixel_size=0.1)

        ag_cone_2D = AcquisitionGeometry.create_Cone2D(source_position=[0,-10],detector_position=[0,10], units='mm')
        ag_cone_2D.set_angles(angles)
        ag_cone_2D.set_panel(20)

        ag_cone_3D = AcquisitionGeometry.create_Cone3D(source_position=[0,-10,0],detector_position=[0,10,0], units='mm')
        ag_cone_3D.set_angles(angles)
        ag_cone_3D.set_panel([20,2], pixel_size=0.1)

        ag = AcquisitionGeometry.create_Parallel3D(detector_position=[0, 150, 0], units='mm').set_panel([20,2], pixel_size=0.1).set_angles(angles)
        
        self.geometry_test_list = [ag_parallel_2D, ag_parallel_3D, ag_cone_2D, ag_cone_3D]


    def fail(self, error_message):
        raise NotImplementedError(error_message)
    
    def out_check(self, processor, data, data_array_index, *args):
        """
        - Test to check the processor gives the same result using the out argument
        - Check the out argument doesn't change the orginal data
        - Check the processor still returns correctly with the out argument
        - Check the processor returns an error if the dtype of out is different to expected
        """
        data_copy=data.copy() #To check that data isn't changed after function calls
        try:
            processor.set_input(data_copy)
            out1 = processor.get_output()
            out2 = out1.copy()
            out2.fill(0)
            out3 = processor.get_output(out=out2)

            # check the processor gives the same result using the out argument
            self.assertDataContainerAllClose(out1, out2, rtol=1e-10, strict=True)
            # check the out argument doesn't change the orginal data
            self.assertDataContainerAllClose(data_copy, data, rtol=1e-10, strict=True)
            # check the processor still returns correctly with the out argument
            self.assertEqual(id(out2), id(out3))
            # check the processor returns an error if the dtype of out is different to expected
            out_wrong_type = out1.copy()
            out_wrong_type.array = numpy.array(out_wrong_type.array, dtype=np.int64)
            with self.assertRaises(TypeError):
                processor.get_output(out=out_wrong_type)

        except NotImplementedError:
            self.fail("out_test test not implemented for  " + processor.__class__.__name__)
        except Exception as e:
            error_message = '\nFor processor: ' + processor.__class__.__name__ + \
            '\nOn data_array index: ' + str(data_array_index) +\
            '\nFor geometry type: \n' + str(data.geometry) 
            raise type(e)(error_message + '\n\n' + str(e))

    def in_place_check(self, processor, data, data_array_index, *args):
        """
        Check the processor gives the correct result if the calculation is performed in place
        i.e. data is passed to the out argument
        if the processor output is a different sized array, the calcualtion
        cannot be perfomed in place, so we expect this to raise a ValueError
        """
        data_copy = data.copy()
        try:
            processor.set_input(data_copy)
            out1 = processor.get_output()

            # Check an error is raised for Processor's that change the dimensions of the data
            if out1.shape != data.shape:
                with self.assertRaises(ValueError):
                    processor.get_output(out=data)
            # Check an error is raised if the Processor returns a different data type (only MaskGenerator at the moment)
            elif isinstance(processor, MaskGenerator):
                    with self.assertRaises(TypeError):
                        processor.get_output(out=data)
            # Otherwise check Processor's work correctly in place
            else:
                processor.get_output(out=data_copy)
                self.assertDataContainerAllClose(out1, data_copy, rtol=1e-10, strict=True)

        except (InPlaceError, NotImplementedError):
            self.fail("in_place_test test not implemented for  " + processor.__class__.__name__)
        except Exception as e:
            error_message = '\nFor processor: ' + processor.__class__.__name__ + \
            '\nOn data_array index: ' + str(data_array_index) +\
            '\nFor geometry type: \n' + str(data.geometry) 
            raise type(e)(error_message + '\n\n' + str(e))

    def test_out(self):
        """
        Tests to check output from Processors, including:
        - Checks that Processors give the same results when the result is obtained
        directly or using the out argument, or in place
        - Checks that the out argument doesn't change the original data
        - Checks that Processors still return the output even when out is used
        All the tests are performed on different data types, geometry and Processors
        Any new Processors should be added to the processor_list
        """

        # perform the tests on different data and geometry types
        for geom in self.geometry_test_list:
            for i, data_array in enumerate(self.data_arrays):
                data=geom.allocate(None)
                if AcquisitionType.DIM2 & geom.dimension:
                    data.fill(data_array)
                else:
                    data.fill(np.repeat(data_array[:,None, :], repeats=2, axis=1))
                # Add new Processors here 
                processor_list = [
                    TransmissionAbsorptionConverter(min_intensity=0.01),
                    AbsorptionTransmissionConverter(),
                    RingRemover(info=False),
                    Slicer(roi={'horizontal':(None,None,None),'angle':(None,None,None)}), 
                    Slicer(roi={'horizontal':(1,3,2),'angle':(None,4,2)}), 
                    Binner(roi={'horizontal':(None,None,None),'angle':(None,None,None)}),
                    Binner(roi={'horizontal':(1,None,2),'angle':(None,4,2)}),
                    Padder(pad_width=0),
                    Padder(pad_width=1),
                    Normaliser(flat_field=data.get_slice(angle=0).as_array()*1, dark_field=data.get_slice(angle=0).as_array()*1e-5),
                    MaskGenerator.median(threshold_factor=3, window=7),
                    Masker.median(mask=data.copy()),
                    PaganinProcessor(),
                    FluxNormaliser(flux=1)
                    ]
                
                for processor in processor_list:
                    self.out_check(processor, data, i)
                    self.in_place_check(processor, data, i)
   
    def test_centre_of_rotation_xcorrelation_out(self):
        """
        Test the output of the centre of rotation xcorrelation processor 
        These tests are performed separately because the processor returns 
        changes to the geometry rather than the data
        """
        # CentreOfRotationCorrector.xcorrelation() works on parallel beam data only
        for geom in self.geometry_test_list[0:2]:
            for data_array in self.data_arrays:
                data=geom.allocate(None)
                try:
                    data.fill(data_array)
                except:
                    data.fill(np.repeat(data_array[:,None, :], repeats=2, axis=1))
                processor = CentreOfRotationCorrector.xcorrelation(ang_tol=180)
                processor.set_input(data)
                out1 = processor.get_output()
                # out_test fails because the processor only updates geometry, this is expected behaviour
                # self.out_test(result, processor, data)
                # test geometry instead
                input = data.copy()
                out2 = 0*(out1.copy())
                try:
                    processor.set_input(input)
                    out3 = processor.get_output(out=out2)
                    
                    numpy.testing.assert_allclose(out1.geometry.config.system.rotation_axis.position, out2.geometry.config.system.rotation_axis.position, err_msg= "Calculation failed using processor.set_input(data), processor.get_output(out=out) where func is  " + processor.__class__.__name__+ ".")
                    numpy.testing.assert_allclose(input.geometry.config.system.rotation_axis.position, data.geometry.config.system.rotation_axis.position,  err_msg= "In case processor.set_input(data), processor.get_output(out=data) where processor is  " + processor.__class__.__name__+ " the input data has been incorrectly affected by the calculation. ")
                    self.assertDataArraysInContainerAllClose(out2, out3,  rtol=1e-5, msg= "In case processor.set_input(data), output=processor.get_output(out=data) where processor is  " + processor.__class__.__name__+ " the processor incorrectly supresses the output. ")

                    processor.set_input(data)
                    processor.get_output(out=data)
                    numpy.testing.assert_array_equal(out1.geometry.config.system.rotation_axis.position, data.geometry.config.system.rotation_axis.position, err_msg= "Calculation failed using processor.set_input(data), processor.get_output(out=data) where func is  " + processor.__class__.__name__+ ".")
                    
                except (InPlaceError, NotImplementedError):
                    print("out_test_for_geometry test not implemented for  " + processor.__class__.__name__)
                    pass

                try:
                    if out1.shape != data.shape:
                        with self.assertRaises(ValueError):
                            processor.get_output(out=data)
                    else:
                        processor.get_output(out=data)
                        self.assertDataArraysInContainerAllClose(out1, data, rtol=1e-5, msg= "In place calculation failed for processor.set_input(data), processor.get_output(out=data) where processor is  " + processor.__class__.__name__+ "." )
                except (InPlaceError, NotImplementedError):
                    print("in_place_test test not implemented for  " + processor.__class__.__name__)
    
    @unittest.skipUnless(has_tigre and has_nvidia, "TIGRE GPU not installed")
    def test_centre_of_rotation_image_sharpness_out(self):
        """
        Test the output of the centre of rotation image_sharpness processor 
        These tests are performed separately because the processor returns 
        changes to the geometry rather than the data
        """
        # Here we need to use real data rather than random data otherwise the processor fails
        data = dataexample.SIMULATED_PARALLEL_BEAM_DATA.get()
        
        processor = Binner(roi={'horizontal':(0,-1,2),'vertical':(0,-1,2), 'angle':(0,-1,2)})
        processor.set_input(data)
        data = processor.get_output()

        processor = CentreOfRotationCorrector.image_sharpness()
        processor.set_input(data)
        out1 = processor.get_output()
        # out_test fails because the processor only updates geometry, this is expected behaviour
        # self.out_test(result, processor, data)
        # test geometry instead
        input = data.copy()
        out2 = 0*(out1.copy())
        try:
            processor.set_input(input)
            out3 = processor.get_output(out=out2)
            
            numpy.testing.assert_allclose(out1.geometry.config.system.rotation_axis.position, out2.geometry.config.system.rotation_axis.position, err_msg= "Calculation failed using processor.set_input(data), processor.get_output(out=out) where func is  " + processor.__class__.__name__+ ".")
            numpy.testing.assert_allclose(input.geometry.config.system.rotation_axis.position, data.geometry.config.system.rotation_axis.position,  err_msg= "In case processor.set_input(data), processor.get_output(out=data) where processor is  " + processor.__class__.__name__+ " the input data has been incorrectly affected by the calculation. ")
            self.assertDataArraysInContainerAllClose(out2, out3,  rtol=1e-5, msg= "In case processor.set_input(data), output=processor.get_output(out=data) where processor is  " + processor.__class__.__name__+ " the processor incorrectly supresses the output. ")

            processor.set_input(data)
            processor.get_output(out=data)
            numpy.testing.assert_array_equal(out1.geometry.config.system.rotation_axis.position, data.geometry.config.system.rotation_axis.position, err_msg= "Calculation failed using processor.set_input(data), processor.get_output(out=data) where func is  " + processor.__class__.__name__+ ".")
            
        except (InPlaceError, NotImplementedError):
            self.fail("out_test_for_geometry test not implemented for  " + processor.__class__.__name__)
            pass

        try:
            if out1.shape != data.shape:
                with self.assertRaises(ValueError):
                    processor.get_output(out=data)
            else:
                processor.get_output(out=data)
                self.assertDataArraysInContainerAllClose(out1, data, rtol=1e-5, msg= "In place calculation failed for processor.set_input(data), processor.get_output(out=data) where processor is  " + processor.__class__.__name__+ "." )
        except (InPlaceError, NotImplementedError):
            self.fail("in_place_test test not implemented for  " + processor.__class__.__name__)


#  Copyright 2018 United Kingdom Research and Innovation
#  Copyright 2018 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
from cil.framework import DataProcessor, ImageData
from cil.framework.labels import AcquisitionDimension, ImageDimension
from cil.plugins.astra.utilities import convert_geometry_to_astra_vec_3D
import astra
from astra import astra_dict, algorithm, data3d
import numpy as np


class AstraBackProjector3D(DataProcessor):

    """
    AstraBackProjector3D configures an ASTRA 3D back projector for GPU.

    Parameters
    ----------

    volume_geometry : ImageGeometry
        A description of the area/volume to reconstruct

    sinogram_geometry : AcquisitionGeometry
        A description of the acquisition data

    """
    def __init__(self,
                 volume_geometry=None,
                 sinogram_geometry=None):
        kwargs = {
                  'volume_geometry'  : volume_geometry,
                  'sinogram_geometry'  : sinogram_geometry,
                  'proj_geom'  : None,
                  'vol_geom'  : None}

        #DataProcessor.__init__(self, **kwargs)
        super(AstraBackProjector3D, self).__init__(**kwargs)

        self.set_ImageGeometry(volume_geometry)
        self.set_AcquisitionGeometry(sinogram_geometry)

        self.vol_geom, self.proj_geom = convert_geometry_to_astra_vec_3D(self.volume_geometry, self.sinogram_geometry)

    def check_input(self, dataset):

        if self.sinogram_geometry.shape != dataset.geometry.shape:
            raise ValueError("Dataset not compatible with geometry used to create the projector")

        return True
    
    def _set_up(self):
        """
        Configure processor attributes that require the data to setup
        Must set _shape_out
        """
        self._shape_out = self.volume_geometry.shape

    def set_ImageGeometry(self, volume_geometry):

        ImageDimension.check_order_for_engine('astra', volume_geometry)

        if len(volume_geometry.dimension_labels) > 3:
            raise ValueError("Supports 2D and 3D data only, got {0}".format(volume_geometry.number_of_dimensions))

        self.volume_geometry = volume_geometry.copy()

    def set_AcquisitionGeometry(self, sinogram_geometry):

        AcquisitionDimension.check_order_for_engine('astra', sinogram_geometry)

        if len(sinogram_geometry.dimension_labels) > 3:
            raise ValueError("Supports 2D and 3D data only, got {0}".format(sinogram_geometry.number_of_dimensions))

        self.sinogram_geometry = sinogram_geometry.copy()

    def process(self, out=None):

        DATA = self.get_input()

        #ASTRA expects a 3D array with shape 1, CIL removes dimensions of len 1
        new_shape_ag = [self.sinogram_geometry.pixel_num_v,self.sinogram_geometry.num_projections,self.sinogram_geometry.pixel_num_h]
        data_temp = DATA.as_array().reshape(new_shape_ag)

        if out is None:
            rec_id, arr_out = astra.create_backprojection3d_gpu(data_temp,
                            self.proj_geom,
                            self.vol_geom)
        else:
            new_shape_ig = [self.volume_geometry.voxel_num_z,self.volume_geometry.voxel_num_y,self.volume_geometry.voxel_num_x]
            new_shape_ig = [x if x>0 else 1 for x in new_shape_ig]
            arr_out = out.as_array().reshape(new_shape_ig)

            rec_id = astra.data3d.link('-vol', self.vol_geom, arr_out)
            self.create_backprojection3d_gpu( data_temp, self.proj_geom, self.vol_geom, False, rec_id)

        # delete the GPU copy
        astra.data3d.delete(rec_id)

        arr_out = np.squeeze(arr_out)

        if out is None:
            out = ImageData(arr_out, deep_copy=False, geometry=self.volume_geometry.copy(), suppress_warning=True)
            return out
        else:
            out.fill(arr_out)

    def create_backprojection3d_gpu(self, data, proj_geom, vol_geom, returnData=True, vol_id=None):

        """
        Call to ASTRA to create a backward projection of an image (3D)

        Parameters
        ----------

        data : numpy.ndarray or int
            Image data or ID.

        proj_geom : dict
            Projection geometry.

        vol_geom : dict
            Volume geometry.

        returnData : bool
            If False, only return the ID of the forward projection.

        vol_id : int, default None
            ID of the np array linked with astra.

        Returns
        -------

        proj_geom : int or (int, numpy.ndarray)
            If ``returnData=False``, returns the ID of the back projection. Otherwise returns a tuple containing the ID of the back projection and the back projection itself.
        """

        if isinstance(data, np.ndarray):
            sino_id = data3d.create('-sino', proj_geom, data)
        else:
            sino_id = data

        if vol_id is None:
            vol_id = data3d.create('-vol', vol_geom, 0)

        cfg = astra_dict('BP3D_CUDA')
        cfg['ProjectionDataId'] = sino_id
        cfg['ReconstructionDataId'] = vol_id
        alg_id = algorithm.create(cfg)
        algorithm.run(alg_id)
        algorithm.delete(alg_id)

        if isinstance(data, np.ndarray):
            data3d.delete(sino_id)

        if vol_id is not None:
            if returnData:
                return vol_id, data3d.get_shared(vol_id)
            else:
                return vol_id
        else:
            if returnData:
                return vol_id, data3d.get(vol_id)
            else:
                return vol_id



#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import unittest
import warnings
from utils import initialise_tests
import numpy as np
from cil.utilities import dataexample
from cil.utilities import noise
from cil.utilities.quality_measures import mse, mae, psnr
from packaging import version
from cil.processors import Slicer
from testclass import CCPiTestClass
if version.parse(np.version.version) >= version.parse("1.13"):
    try:
        from skimage.metrics import mean_squared_error, peak_signal_noise_ratio
        has_skimage = True
    except ImportError as ie:
            has_skimage = False
else:
    has_skimage = False

initialise_tests()

@unittest.skipIf((not has_skimage) , "Skip test with has_skimage {}".format( has_skimage))
class TestQualityMeasures(CCPiTestClass):

    def setUp(self):
        if has_skimage:

            id_coins = dataexample.CAMERA.get()

            id_coins_noisy = noise.gaussian(id_coins, var=0.05, seed=10)

            ig = id_coins.geometry.copy()
            dc1 = ig.allocate('random')
            dc2 = ig.allocate('random')

            self.dc1 = dc1
            self.dc2 = dc2

            self.mask=ig.allocate(0)
            self.mask.array[:50,:50]=1

            self.bool_mask=self.mask.array.astype('bool')

            self.id_coins = id_coins
            self.id_coins_noisy = id_coins_noisy

            roi = {'horizontal_x':(0,50,1),'horizontal_y':(0,50,1)}
            processor = Slicer(roi)
            processor.set_input(id_coins)
            self.id_coins_sliced= processor.get_output()
            processor = Slicer(roi)
            processor.set_input(id_coins_noisy )
            self.id_coins_noisy_sliced= processor.get_output()


    def test_mse1(self):
        res1 = mse(self.id_coins, self.id_coins_noisy)
        res2 = mean_squared_error(self.id_coins.as_array(), self.id_coins_noisy.as_array())
        np.testing.assert_almost_equal(res1, res2, decimal=5)


    def test_mse2(self):
        res1 = mse(self.dc1, self.dc2)
        res2 = mean_squared_error(self.dc1.as_array(), self.dc2.as_array())
        np.testing.assert_almost_equal(res1, res2, decimal=5)

    def test_psnr1(self):
        res1 = psnr(self.id_coins, self.id_coins_noisy, data_range = self.dc1.max())
        res2 = peak_signal_noise_ratio(self.id_coins.as_array(), self.id_coins_noisy.as_array())
        np.testing.assert_almost_equal(res1, res2, decimal=3)

    def test_psnr2_default_data_range(self):
        res1 = psnr(self.id_coins, self.id_coins_noisy)
        res2 = peak_signal_noise_ratio(self.id_coins.as_array(), self.id_coins_noisy.as_array())
        np.testing.assert_almost_equal(res1, res2, decimal=3)


    def test_psnr2(self):
        res1 = psnr(self.dc1, self.dc2, data_range = self.dc1.max())
        res2 = peak_signal_noise_ratio(self.dc1.as_array(), self.dc2.as_array())
        np.testing.assert_almost_equal(res1, res2, decimal=3)

    def test_mse_mask(self):
        res1 = mse(self.id_coins_sliced, self.id_coins_noisy_sliced)
        res2 = mse(self.id_coins, self.id_coins_noisy, mask=self.mask.array)
        np.testing.assert_almost_equal(res1, res2, decimal=3)

    def test_mse_bool_mask(self):
        res1 = mse(self.id_coins_sliced, self.id_coins_noisy_sliced)
        res2 = mse(self.id_coins, self.id_coins_noisy, mask=self.bool_mask)
        np.testing.assert_almost_equal(res1, res2, decimal=3)

    def test_mse_data_container_mask(self):
        res1 = mse(self.id_coins_sliced, self.id_coins_noisy_sliced)
        res2 = mse(self.id_coins, self.id_coins_noisy, mask=self.mask)
        np.testing.assert_almost_equal(res1, res2, decimal=3)

    def test_psnr_mask(self):
        res1 = psnr(self.id_coins_sliced, self.id_coins_noisy_sliced)
        res2 = psnr(self.id_coins, self.id_coins_noisy, mask=self.mask)
        np.testing.assert_almost_equal(res1, res2, decimal=3)

    def test_mae_mask(self):
        res1 = mae(self.id_coins_sliced, self.id_coins_noisy_sliced)
        res2 = mae(self.id_coins, self.id_coins_noisy, mask=self.mask)
        np.testing.assert_almost_equal(res1, res2, decimal=3)

    def test_infinite_psnr(self):
        with warnings.catch_warnings(record=True) as w:
            self.assertEqual(psnr(self.id_coins_sliced, self.id_coins_sliced), np.inf)
            assert issubclass(w[-1].category, RuntimeWarning)
